from abc import ABC, abstractmethod
import boto3
from spotty.providers.aws.aws_resources.image import Image
from spotty.providers.aws.aws_resources.subnet import Subnet
from spotty.providers.aws.aws_resources.vpc import Vpc
from spotty.providers.aws.config.instance_config import InstanceConfig
from spotty.providers.aws.deployment.project_resources.key_pair import KeyPairResource


class AbstractAwsDeployment(ABC):

    def __init__(self, project_name: str, instance_config: InstanceConfig):
        self._project_name = project_name
        self._instance_config = instance_config
        self._ec2 = boto3.client('ec2', region_name=instance_config.region)

    @property
    def instance_config(self) -> InstanceConfig:
        return self._instance_config

    @property
    @abstractmethod
    def ec2_instance_name(self) -> str:
        """Name for EC2 instance."""
        raise NotImplementedError

    @property
    def key_pair(self) -> KeyPairResource:
        return KeyPairResource(self._project_name, self.instance_config.region)

    def get_vpc_id(self) -> str:
        """Returns VPC ID that should be used for deployment."""
        if self.instance_config.subnet_id:
            vpc_id = Subnet.get_by_id(self._ec2, self.instance_config.subnet_id).vpc_id
        else:
            default_vpc = Vpc.get_default_vpc(self._ec2)
            if not default_vpc:
                raise ValueError('Default VPC not found')

            vpc_id = default_vpc.vpc_id

        return vpc_id

    def get_ami(self) -> Image:
        if self.instance_config.ami_id:
            image = Image.get_by_id(self._ec2, self.instance_config.ami_id)
        else:
            image = Image.get_by_name(self._ec2, self.instance_config.ami_name)

        return image



from spotty.commands.abstract_provider_command import AbstractProviderCommand
from spotty.providers.aws.commands.clean_logs import CleanLogsCommand
from spotty.providers.aws.commands.create_ami import CreateAmiCommand
from spotty.providers.aws.commands.delete_ami import DeleteAmiCommand
from spotty.providers.aws.commands.spot_prices import SpotPricesCommand


class AwsCommand(AbstractProviderCommand):

    name = 'aws'
    description = 'AWS commands'
    commands = [
        CreateAmiCommand,
        DeleteAmiCommand,
        SpotPricesCommand,
        CleanLogsCommand,
    ]



from time import sleep
from botocore.exceptions import EndpointConnectionError, ClientError
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class Stack(object):

    def __init__(self, cf, stack_info):
        self._cf = cf
        self._stack_info = stack_info

    @staticmethod
    def get_by_name(cf, stack_name: str):
        """Returns a Stack by its name."""
        try:
            res = cf.describe_stacks(StackName=stack_name)
        except ClientError as e:
            # ignore an exception if it raised because the stack doesn't exist
            error_code = e.response.get('Error', {}).get('Code')
            if error_code != 'ValidationError':
                raise e

            res = {'Stacks': []}

        if not len(res['Stacks']):
            return None

        return Stack(cf, res['Stacks'][0])

    @staticmethod
    def create_stack(cf, *args, **kwargs):
        res = cf.create_stack(*args, **kwargs)
        return Stack(cf, res)

    @staticmethod
    def update_stack(cf, *args, **kwargs):
        res = cf.update_stack(*args, **kwargs)
        return Stack(cf, res)

    @property
    def stack_id(self) -> str:
        return self._stack_info['StackId']

    @property
    def name(self) -> str:
        return self._stack_info['StackName']

    @property
    def status(self) -> str:
        return self._stack_info['StackStatus']

    @property
    def outputs(self) -> str:
        return self._stack_info['Outputs']

    def delete(self):
        return self._cf.delete_stack(StackName=self.stack_id)

    def wait_stack_created(self, delay=30):
        waiter = self._cf.get_waiter('stack_create_complete')
        waiter.wait(StackName=self.stack_id, WaiterConfig={'Delay': delay})

    def wait_stack_updated(self, delay=30):
        waiter = self._cf.get_waiter('stack_update_complete')
        waiter.wait(StackName=self.stack_id, WaiterConfig={'Delay': delay})

    def wait_stack_deleted(self, delay=30):
        waiter = self._cf.get_waiter('stack_delete_complete')
        waiter.wait(StackName=self.stack_id, WaiterConfig={'Delay': delay})

    def wait_status_changed(self, waiting_status, resource_messages, resource_success_status,
                            output: AbstractOutputWriter, delay=5):
        current_status = waiting_status
        stack = None

        resource_messages = iter(resource_messages) if resource_messages else None
        resource_name = None

        while current_status == waiting_status:
            sleep(delay)

            # display resource creation progress
            if resource_messages:
                try:
                    stack_resources = self._cf.list_stack_resources(StackName=self.stack_id)
                except EndpointConnectionError:
                    output.write('Connection problem')
                    continue

                resource_statuses = dict([(row['LogicalResourceId'], row['ResourceStatus'])
                                          for row in stack_resources['StackResourceSummaries']])

                while (resource_name is None) or (resource_name and
                                                  resource_statuses.get(resource_name, '') == resource_success_status):
                    (resource_name, resource_msg) = next(resource_messages, (False, False))
                    if resource_name:
                        output.write('- %s...' % resource_msg)

            # get the latest status of the stack
            try:
                stack = self.get_by_name(self._cf, self.stack_id)
            except EndpointConnectionError:
                output.write('Connection problem')
                continue

            current_status = stack.status

        return stack



from importlib import import_module
from spotty.config.project_config import ProjectConfig
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class InstanceManagerFactory(object):

    SUPPORTED_PROVIDERS = ['aws']

    @classmethod
    def get_instance(cls, project_config: ProjectConfig, instance_config: dict) -> AbstractInstanceManager:
        provider_name = instance_config['provider']
        if provider_name not in cls.SUPPORTED_PROVIDERS:
            raise ValueError('Provider "%s" is not supported' % provider_name)

        # get Instance Manger class for the provider
        InstanceManagerClass = getattr(import_module('spotty.providers.%s.instance_manager' % provider_name),
                                       'InstanceManager')

        return InstanceManagerClass(project_config, instance_config)



import boto3
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.aws.aws_resources.instance import Instance
from spotty.providers.aws.aws_resources.stack import Stack


class InstanceStackResource(object):

    def __init__(self, project_name: str, instance_name: str, region: str):
        self._cf = boto3.client('cloudformation', region_name=region)
        self._ec2 = boto3.client('ec2', region_name=region)
        self._region = region
        self._stack_name = 'spotty-instance-%s-%s' % (project_name.lower(), instance_name.lower())

    @property
    def name(self):
        return self._stack_name

    def get_instance(self):
        return Instance.get_by_stack_name(self._ec2, self.name)

    def create_or_update_stack(self, template: str, parameters: dict, output: AbstractOutputWriter):
        """Runs CloudFormation template."""

        # delete the stack if it exists
        stack = Stack.get_by_name(self._cf, self._stack_name)
        if stack:
            self.delete_stack(output)

        # create new stack
        stack = Stack.create_stack(
            cf=self._cf,
            StackName=self._stack_name,
            TemplateBody=template,
            Parameters=[{'ParameterKey': key, 'ParameterValue': value} for key, value in parameters.items()],
            Capabilities=['CAPABILITY_IAM'],
            OnFailure='DO_NOTHING',
        )

        output.write('Waiting for the stack to be created...')

        resource_messages = [
            ('Instance', 'launching the instance'),
            ('DockerReadyWaitCondition', 'waiting for the Docker container to be ready'),
        ]

        # wait for the stack to be created
        with output.prefix('  '):
            stack = stack.wait_status_changed(waiting_status='CREATE_IN_PROGRESS',
                                              resource_messages=resource_messages,
                                              resource_success_status='CREATE_COMPLETE', output=output)

        if stack.status != 'CREATE_COMPLETE':
            raise ValueError('Stack "%s" was not created.\n'
                             'Please, see CloudFormation logs for the details.'
                             % self._stack_name)

        return stack

    def delete_stack(self, output: AbstractOutputWriter, no_wait=False):
        stack = Stack.get_by_name(self._cf, self._stack_name)
        if not stack:
            return

        if not no_wait:
            output.write('Waiting for the stack to be deleted...')

        # delete the stack
        try:
            stack.delete()
            if not no_wait:
                stack.wait_stack_deleted()
        except Exception as e:
            raise ValueError('Stack "%s" was not deleted. Error: %s\n'
                             'See CloudFormation logs for details.' % (self._stack_name, str(e)))



from argparse import Namespace
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class StopCommand(AbstractConfigCommand):

    name = 'stop'
    description = 'Terminate running instance and apply deletion policies for the volumes'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):

        instance_manager.stop(output)

        output.write('\n'
                     '----------------------------------\n'
                     'Instance was successfully deleted.\n'
                     '----------------------------------')



class Subnet(object):

    def __init__(self, ec2, subnet_info):
        self._ec2 = ec2
        self._subnet_info = subnet_info

    @staticmethod
    def get_by_id(ec2, subnet_id: str):
        """Returns a subnet by its ID."""
        res = ec2.describe_subnets(Filters=[
            {'Name': 'subnet-id', 'Values': [subnet_id]},
        ])

        if not len(res['Subnets']):
            return None

        return Subnet(ec2, res['Subnets'][0])

    @staticmethod
    def get_default_subnets(ec2):
        res = ec2.describe_subnets(Filters=[
            {'Name': 'defaultForAz', 'Values': ['true']},
        ])

        subnets = [Subnet(ec2, subnet_info) for subnet_info in res['Subnets']]

        return subnets

    @property
    def availability_zone(self) -> str:
        return self._subnet_info['AvailabilityZone']

    @property
    def vpc_id(self) -> str:
        return self._subnet_info['VpcId']



from argparse import Namespace, ArgumentParser
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class StartCommand(AbstractConfigCommand):

    name = 'start'
    description = 'Run a spot instance, sync the project and start the Docker container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('--dry-run', action='store_true', help='Displays the steps that would be performed '
                                                                   'using the specified command without actually '
                                                                   'running them')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # start the instance
        dry_run = args.dry_run
        with output.prefix('[dry-run] ' if dry_run else ''):
            instance_manager.start(output, dry_run)

        if not dry_run:
            instance_name = ''
            if len(instance_manager.project_config.instances) > 1:
                instance_name = ' ' + instance_manager.instance_config.name

            output.write('\nThe instance was successfully started.\n'
                         '\n%s\n'
                         '\nUse the "spotty ssh%s" command to connect to the Docker container.\n'
                         % (instance_manager.status_text, instance_name))



from argparse import Namespace
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class StatusCommand(AbstractConfigCommand):

    name = 'status'
    description = 'Print information about the instance'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        output.write(instance_manager.status_text)



from argparse import Namespace
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.abstract_instance_manager import AbstractInstanceManager
from spotty.providers.aws.instance_manager import InstanceManager


class DeleteAmiCommand(AbstractConfigCommand):

    name = 'delete-ami'
    description = 'Delete AMI with NVIDIA Docker'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that it's an AWS instance
        if not isinstance(instance_manager, InstanceManager):
            raise ValueError('Instance "%s" is not an AWS instance.' % instance_manager.instance_config.name)

        instance_manager.ami_deployment.delete(output)



from argparse import ArgumentParser, Namespace
import subprocess
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.helpers.ssh import get_ssh_command
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class SshCommand(AbstractConfigCommand):

    name = 'ssh'
    description = 'Connect to the running Docker container or to the instance itself'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-H', '--host-os', action='store_true', help='Connect to the host OS instead of the Docker '
                                                                         'container')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-l', '--list-sessions', action='store_true', help='List all tmux sessions managed by the '
                                                                               'instance')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        if args.list_sessions:
            remote_cmd = ['tmux', 'ls', ';', 'echo', '']
        else:
            # tmux session name
            session_name = args.session_name
            if not session_name:
                session_name = 'spotty-ssh-host-os' if args.host_os else 'spotty-ssh-container'

            # a command to connect to the host OS or to the container
            remote_cmd = ['tmux', 'new', '-s', session_name, '-A']
            if not args.host_os:
                # connect to the container or keep the tmux window in case of a failure
                container_cmd = subprocess.list2cmdline(['sudo', '/tmp/spotty/instance/scripts/container_bash.sh'])
                tmux_cmd = '%s || tmux set remain-on-exit on' % container_cmd
                remote_cmd += [tmux_cmd]

        remote_cmd = subprocess.list2cmdline(remote_cmd)

        # connect to the instance
        ssh_command = get_ssh_command(instance_manager.ip_address, instance_manager.ssh_port,
                                      instance_manager.ssh_user, instance_manager.ssh_key_path, remote_cmd)
        subprocess.call(ssh_command)



import base64
import subprocess


def get_ssh_command(host: str, port: int, user: str, key_path: str, remote_cmd: str, quiet: bool = False) -> list:
    ssh_command = ['ssh', '-ti', key_path, '-o', 'StrictHostKeyChecking no']

    if port != 22:
        ssh_command += ['-p', str(port)]

    if quiet:
        ssh_command += ['-q']

    ssh_command += ['%s@%s' % (user, host), remote_cmd]

    return ssh_command


def run_script(host: str, port: int, user: str, key_path: str, script_name: str, script_content: str,
               tmux_session_name: str, restart: bool = False, logging: bool = False):
    # encode the script content to base64
    script_base64 = base64.b64encode(script_content.encode('utf-8')).decode('utf-8')

    # a remote path where the script will be uploaded
    instance_script_path = '/tmp/spotty/container/scripts/run/%s.sh' % script_name
    container_script_path = '/tmp/scripts/run/%s.sh' % script_name

    # command to attach user to existing tmux session
    attach_session_cmd = subprocess.list2cmdline(['tmux', 'attach', '-t', tmux_session_name, '>', '/dev/null',
                                                  '2>&1'])

    # command to kill session in case of a restart
    kill_session_cmd = subprocess.list2cmdline(['tmux', 'kill-session', '-t', tmux_session_name, '>', '/dev/null',
                                                '2>&1'])

    # command to upload user script to the instance
    upload_script_cmd = subprocess.list2cmdline(['echo', script_base64, '|', 'base64', '-d', '>', instance_script_path])

    # log the script outputs to the file
    log_cmd = ['2>&1', '|', 'tee', '/var/log/spotty/run/%s-`date +%%s`.log' % script_name] if logging else []

    # command to run user script inside the docker container
    docker_cmd = subprocess.list2cmdline(['sudo', '/tmp/spotty/instance/scripts/container_bash.sh', '-xe',
                                          container_script_path] + log_cmd)

    # command to create new tmux session and run user script
    new_session_cmd = subprocess.list2cmdline(['tmux', 'new', '-s', tmux_session_name, '-n', script_name,
                                               'tmux set remain-on-exit on && %s' % docker_cmd])

    if restart:
        # composition of the commands: killing the script session if it already exists, then uploading the script
        # to the instance, creating new tmux session and running the script inside the Docker container
        remote_cmd = '%s; (%s && %s)' % (kill_session_cmd, upload_script_cmd, new_session_cmd)
    else:
        # composition of the commands: trying to attach the user to the existing tmux session. If it doesn't exist,
        # uploading the user script to the instance, creating new tmux session and running that script
        # inside the Docker container
        remote_cmd = '%s || (%s && %s)' % (attach_session_cmd, upload_script_cmd, new_session_cmd)

    # connect to the instance and run the command
    ssh_command = get_ssh_command(host, port, user, key_path, remote_cmd)
    subprocess.call(ssh_command)



from abc import ABC, abstractmethod
from contextlib import contextmanager


class AbstractOutputWriter(ABC):

    def __init__(self):
        self._prefix = ''

    @abstractmethod
    def _write(self, msg: str):
        raise NotImplementedError

    def write(self, msg: str = ''):
        msg = '\n'.join([self._prefix + line for line in msg.split('\n')])
        self._write(msg)

    @contextmanager
    def prefix(self, prefix):
        self._prefix += prefix
        yield
        self._prefix = self._prefix[:-len(prefix)]



import os
import boto3
from spotty.configuration import get_spotty_keys_dir


class KeyPairResource(object):

    def __init__(self, project_name: str, region: str):
        self._ec2 = boto3.client('ec2', region_name=region)
        self._region = region
        self._key_name = 'spotty-key-%s-%s' % (project_name.lower(), region)

    @property
    def key_path(self):
        return os.path.join(get_spotty_keys_dir(), self._key_name)

    def get_or_create_key(self, dry_run=False):
        if dry_run:
            return self._key_name

        key_path = self.key_path
        key_file_exists = os.path.isfile(key_path)
        ec2_key_exists = self._ec2_key_exists()

        if not ec2_key_exists or not key_file_exists:
            # remove key from AWS (key file not found)
            if ec2_key_exists:
                self._ec2.delete_key_pair(KeyName=self._key_name)

            # create new key
            res = self._ec2.create_key_pair(KeyName=self._key_name)

            # save the key to spotty directory
            with open(key_path, 'w') as f:
                f.write(res['KeyMaterial'])

            os.chmod(key_path, 0o600)

        return self._key_name

    def delete_key(self):
        # delete EC2 Key Pair
        if self._ec2_key_exists():
            self._ec2.delete_key_pair(KeyName=self._key_name)

        # delete the key file
        key_path = self.key_path
        if os.path.isfile(key_path):
            os.unlink(key_path)

    def _ec2_key_exists(self):
        res = self._ec2.describe_key_pairs(Filters=[{'Name': 'key-name', 'Values': [self._key_name]}])
        if 'KeyPairs' not in res:
            return False

        if len(res['KeyPairs']) > 1:
            raise ValueError('Several keys with the name "%s" found.' % self._key_name)

        return bool(res['KeyPairs'])



import subprocess
from spotty.helpers.ssh import get_ssh_command
from spotty.providers.aws.helpers.aws_cli import AwsCli


def get_tmp_instance_s3_path(bucket_name, instance_name):
    return 's3://%s/download/instance-%s' % (bucket_name, instance_name)


def upload_from_instance_to_s3(download_filters: list, host: str, port: int, user: str, key_path: str,
                               dry_run: bool = False):
    """Uploads files from the running instance to the S3 bucket.

    It uses a temporary S3 directory that is unique for the instance. This
    directory keeps downloaded from the instance files in order to sync only
    changed files with local, not all of them every download).
    """

    # "sudo" should be called with the "-i" flag to use the root environment, so aws-cli will read
    # the config file from the root home directory
    args = ['sudo', '-i', '/tmp/spotty/instance/scripts/upload_files.sh']
    args += AwsCli.get_s3_sync_arguments(filters=download_filters, delete=True, quote=True, dry_run=dry_run)

    if not dry_run:
        args += ['>', '/dev/null']

    remote_cmd = subprocess.list2cmdline(args)

    # connect to the instance and run the remote command
    ssh_command = get_ssh_command(host, port, user, key_path, remote_cmd, quiet=not dry_run)
    subprocess.call(ssh_command)


def download_from_s3_to_local(bucket_name: str, instance_name: str, project_dir: str, region: str,
                              download_filters: list, dry_run: bool = False):
    """Downloads files from a temporary S3 directory to local."""
    AwsCli(region=region).s3_sync(get_tmp_instance_s3_path(bucket_name, instance_name), project_dir,
                                  filters=download_filters, exact_timestamp=True, capture_output=False,
                                  dry_run=dry_run)



class InstanceNotRunningError(Exception):
    def __init__(self, instance_name: str):
        super().__init__('Instance "%s" is not running.\n'
                         'Use the "spotty start %s" command to start the instance.'
                         % (instance_name, instance_name))



from datetime import datetime
from spotty.providers.aws.helpers.spot_prices import get_current_spot_price


class Instance(object):

    def __init__(self, ec2, instance_info):
        self._ec2 = ec2
        self._instance_info = instance_info

    @staticmethod
    def get_by_stack_name(ec2, stack_name):
        """Returns the running instance by its stack name
           or None if the instance is not running.
        """
        res = ec2.describe_instances(Filters=[
            {'Name': 'tag:aws:cloudformation:stack-name', 'Values': [stack_name]},
            {'Name': 'instance-state-name', 'Values': ['running']},
        ])

        if len(res['Reservations']) > 1:
            raise ValueError('Several running instances for the stack "%s" are found.' % stack_name)

        if not len(res['Reservations']):
            return None

        return Instance(ec2, res['Reservations'][0]['Instances'][0])

    def terminate(self):
        return self._ec2.terminate_instances(InstanceIds=[self.instance_id])

    def wait_instance_terminated(self):
        waiter = self._ec2.get_waiter('instance_terminated')
        waiter.wait(InstanceIds=[self.instance_id])

    @property
    def instance_id(self):
        return self._instance_info['InstanceId']

    @property
    def public_ip_address(self) -> str:
        return self._instance_info.get('PublicIpAddress', None)

    @property
    def state(self) -> str:
        return self._instance_info['State']['Name']

    @property
    def instance_type(self) -> str:
        return self._instance_info['InstanceType']

    @property
    def availability_zone(self) -> str:
        return self._instance_info['Placement']['AvailabilityZone']

    @property
    def launch_time(self) -> datetime:
        return self._instance_info['LaunchTime']

    @property
    def lifecycle(self) -> str:
        return self._instance_info['InstanceLifecycle']

    def get_spot_price(self):
        """Get current Spot Instance price for this instance."""
        return get_current_spot_price(self._ec2, self.instance_type, self.availability_zone)



from argparse import ArgumentParser
import pkg_resources


def add_subparsers(parser: ArgumentParser, command_classes: list):
    """Adds commands to the parser."""
    subparsers = parser.add_subparsers()
    for command_class in command_classes:
        command = command_class()
        subparser = subparsers.add_parser(command.name, help=command.description)
        subparser.set_defaults(command=command, parser=subparser)
        command.configure(subparser)


def get_custom_commands():
    """Returns custom commands that integrated through entry points."""
    return [entry_point.load() for entry_point in pkg_resources.iter_entry_points('spotty.commands')]



import time


class Snapshot(object):

    def __init__(self, ec2, snapshot_info):
        self._ec2 = ec2
        self._snapshot_info = snapshot_info

    @staticmethod
    def get_by_name(ec2, snapshot_name: str):
        """Returns a snapshot by its name."""
        res = ec2.describe_snapshots(Filters=[
            {'Name': 'tag:Name', 'Values': [snapshot_name]},
        ])

        if len(res['Snapshots']) > 1:
            raise ValueError('Several snapshots with Name=%s found.' % snapshot_name)

        if not len(res['Snapshots']):
            return None

        return Snapshot(ec2, res['Snapshots'][0])

    @property
    def name(self) -> str:
        snapshot_name = [tag['Value'] for tag in self._snapshot_info['Tags'] if tag['Key'] == 'Name']
        if not snapshot_name:
            return ''

        return snapshot_name[0]

    @property
    def snapshot_id(self):
        return self._snapshot_info['SnapshotId']

    @property
    def size(self) -> int:
        return self._snapshot_info['VolumeSize']

    @property
    def creation_time(self) -> int:
        return int(time.mktime(self._snapshot_info['StartTime'].timetuple()))

    def rename(self, new_name):
        return self._ec2.create_tags(Resources=[self.snapshot_id],
                                     Tags=[{'Key': 'Name', 'Value': new_name}])

    def delete(self):
        return self._ec2.delete_snapshot(SnapshotId=self.snapshot_id)

    def wait_snapshot_completed(self):
        waiter = self._ec2.get_waiter('snapshot_completed')
        waiter.wait(SnapshotIds=[self.snapshot_id])



from argparse import ArgumentParser, Namespace
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.errors.instance_not_running import InstanceNotRunningError
from spotty.helpers.run import parse_parameters, render_script
from spotty.helpers.ssh import run_script
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class RunCommand(AbstractConfigCommand):

    name = 'run'
    description = 'Run a script from the configuration file inside the Docker container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-S', '--sync', action='store_true', help='Sync the project before running the script')
        parser.add_argument('-l', '--logging', action='store_true', help='Log the script outputs to the file')
        parser.add_argument('-r', '--restart', action='store_true',
                            help='Restart the script (kills previous session if it exists)')
        parser.add_argument('script_name', metavar='SCRIPT_NAME', type=str, help='Script name')
        parser.add_argument('-p', '--parameters', metavar='PARAMETER=VALUE', nargs='*', type=str, default=[],
                            help='Script parameters')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that the script exists
        script_name = args.script_name
        scripts = instance_manager.project_config.scripts
        if script_name not in scripts:
            raise ValueError('Script "%s" is not defined in the configuration file.' % script_name)

        # replace script parameters
        params = parse_parameters(args.parameters)
        script_content = render_script(scripts[script_name], params)

        # check that the instance is started
        if not instance_manager.is_running():
            raise InstanceNotRunningError(instance_manager.instance_config.name)

        # sync the project with the instance
        if args.sync:
            instance_manager.sync(output)

        # tmux session name
        session_name = args.session_name if args.session_name else 'spotty-script-%s' % script_name

        # run the script on the instance
        run_script(host=instance_manager.ip_address,
                   port=instance_manager.ssh_port,
                   user=instance_manager.ssh_user,
                   key_path=instance_manager.ssh_key_path,
                   script_name=script_name,
                   script_content=script_content,
                   tmux_session_name=session_name,
                   restart=args.restart,
                   logging=args.logging)



import boto3
import re
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.utils import random_string


class BucketResource(object):

    def __init__(self, project_name: str, region: str):
        self._s3 = boto3.client('s3', region_name=region)
        self._region = region
        self._bucket_prefix = 'spotty-%s' % project_name.lower()

    def _find_bucket(self):
        res = self._s3.list_buckets()
        regex = re.compile('-'.join([self._bucket_prefix, '[a-z0-9]{12}', self._region]))
        buckets = [bucket['Name'] for bucket in res['Buckets'] if regex.match(bucket['Name']) is not None]

        if len(buckets) > 1:
            raise ValueError('Found several buckets in the same region: %s.' % ', '.join(buckets))

        bucket_name = buckets[0] if len(buckets) else False

        return bucket_name

    def get_or_create_bucket(self, output: AbstractOutputWriter, dry_run=False):
        bucket_name = self._find_bucket()
        if not bucket_name:
            bucket_name = '-'.join([self._bucket_prefix, random_string(12), self._region])
            if not dry_run:
                self._s3.create_bucket(ACL='private', Bucket=bucket_name,
                                       CreateBucketConfiguration={'LocationConstraint': self._region})
            output.write('Bucket "%s" was created.' % bucket_name)

        return bucket_name

    def delete_bucket(self):
        pass



import os
import random
import string
import errno


def package_dir(path: str = ''):
    """Returns an absolute path to the "spotty" package directory.

    Args:
        path: A relative path to add to the package path.

    """
    res_path = os.path.dirname(os.path.abspath(__file__))
    if path:
        res_path = os.path.join(res_path, path)

    return res_path


def check_path(path):
    """Creates a directory if it doesn't exist."""
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise


def random_string(length: int, chars: str = string.ascii_lowercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(length))


def filter_list(list_of_dicts, key_name, value):
    return [row for row in list_of_dicts if row[key_name] == value]


def render_table(table: list, separate_title=False):
    column_lengths = [max([len(str(row[i])) for row in table]) for i in range(len(table[0]))]
    row_separator = '+-%s-+' % '-+-'.join(['-' * col_length for col_length in column_lengths])
    title_separator = '+=%s=+' % '=+='.join(['=' * col_length for col_length in column_lengths])

    lines = [row_separator]
    for i, row in enumerate(table):
        line = '| %s |' % ' | '.join([str(val).ljust(col_length) for val, col_length in zip(row, column_lengths)])
        lines.append(line)
        lines.append(title_separator if separate_title and not i else row_separator)

    return '\n'.join(lines)



from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.errors.instance_not_running import InstanceNotRunningError
from spotty.providers.aws.config.instance_config import InstanceConfig
from spotty.providers.aws.deployment.ami_deployment import AmiDeployment
from spotty.providers.aws.deployment.instance_deployment import InstanceDeployment
from spotty.providers.aws.errors.ami_not_found import AmiNotFoundError
from spotty.providers.aws.helpers.download import download_from_s3_to_local, upload_from_instance_to_s3
from spotty.providers.aws.helpers.sync import sync_project_with_s3, sync_instance_with_s3
from spotty.providers.abstract_instance_manager import AbstractInstanceManager
from spotty.utils import render_table


class InstanceManager(AbstractInstanceManager):

    @property
    def instance_deployment(self) -> InstanceDeployment:
        """Returns an instance deployment manager."""
        return InstanceDeployment(self.project_config.project_name, self.instance_config)

    @property
    def ami_deployment(self) -> AmiDeployment:
        """Returns an AMI deployment manager."""
        return AmiDeployment(self.project_config.project_name, self.instance_config)

    def _get_instance_config(self, config: dict) -> InstanceConfig:
        """Validates the instance config and returns an InstanceConfig object."""
        return InstanceConfig(config)

    @property
    def instance_config(self) -> InstanceConfig:
        """This property is redefined just for a correct type hinting."""
        return self._instance_config

    def is_running(self):
        """Checks if the instance is running."""
        return bool(self.instance_deployment.get_instance())

    def start(self, output: AbstractOutputWriter, dry_run=False):
        deployment = self.instance_deployment

        if not dry_run:
            # check if the instance is already running
            instance = deployment.get_instance()
            if instance:
                print('Instance is already running. Are you sure you want to restart it?')
                res = input('Type "y" to confirm: ')
                if res != 'y':
                    raise ValueError('The operation was cancelled.')

                # terminating the instance to make EBS volumes available
                output.write('Terminating the instance...')
                instance.terminate()
                instance.wait_instance_terminated()

            # check that the AMI exists
            if not deployment.get_ami():
                print('The AMI "%s" doesn\'t exist. Do you want to create it?'
                      % self.instance_config.ami_name)
                res = input('Type "y" to confirm: ')
                if res == 'y':
                    # create an AMI
                    self.ami_deployment.deploy(False, output)
                    output.write()
                else:
                    raise AmiNotFoundError(self.instance_config.ami_name)

        # deploy the instance
        deployment.deploy(self.project_config, output, dry_run=dry_run)

    def stop(self, output: AbstractOutputWriter):
        # terminate the instance
        instance = self.instance_deployment.get_instance()
        if instance:
            output.write('Terminating the instance...')
            instance.terminate()
            instance.wait_instance_terminated()
        else:
            output.write('The instance is already terminated.')

        # delete the stack
        self.instance_deployment.stack.delete_stack(output, no_wait=True)

        output.write('Applying deletion policies for the volumes...')

        # apply deletion policies for the volumes
        with output.prefix('  '):
            self.instance_deployment.apply_deletion_policies(output)

    def clean(self, output: AbstractOutputWriter):
        pass

    def sync(self, output: AbstractOutputWriter, dry_run=False):
        # create or get existing bucket for the project
        bucket_name = self.instance_deployment.bucket.get_or_create_bucket(output, dry_run)

        # sync the project with S3 bucket
        output.write('Syncing the project with S3 bucket...')
        sync_project_with_s3(self.project_config.project_dir, bucket_name, self.instance_config.region,
                             self.project_config.sync_filters, dry_run=dry_run)

        if not dry_run:
            # sync S3 with the instance
            output.write('Syncing S3 bucket with the instance...')
            sync_instance_with_s3(self.project_config.sync_filters, self.ip_address, self.ssh_port, self.ssh_user,
                                  self.ssh_key_path)

    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):
        # create or get existing bucket for the project
        bucket_name = self.instance_deployment.bucket.get_or_create_bucket(output, dry_run)

        # sync files from the instance to a temporary S3 directory
        output.write('Uploading files from the instance to S3 bucket...')
        upload_from_instance_to_s3(download_filters, self.ip_address, self.ssh_port, self.ssh_user, self.ssh_key_path,
                                   dry_run=dry_run)

        # sync the project with the S3 bucket
        output.write('Downloading files from S3 bucket to local...')
        download_from_s3_to_local(bucket_name, self.instance_config.name, self.project_config.project_dir,
                                  self.instance_config.region, download_filters, dry_run=dry_run)

    @property
    def status_text(self):
        instance = self.instance_deployment.get_instance()
        if not instance:
            raise InstanceNotRunningError(self.instance_config.name)

        table = [
            ('Instance State', instance.state),
            ('Instance Type', instance.instance_type),
            ('Availability Zone', instance.availability_zone),
        ]

        if instance.public_ip_address:
            table.append(('Public IP Address', instance.public_ip_address))

        table.append(('Launch Time', instance.launch_time.strftime('%Y-%m-%d %H:%M:%S')))

        if instance.lifecycle == 'spot':
            spot_price = instance.get_spot_price()
            table.append(('Purchasing Option', 'Spot Instance'))
            table.append(('Spot Instance Price', '$%.04f' % spot_price))
        else:
            table.append(('Purchasing Option', 'On-Demand Instance'))

        return render_table(table)

    @property
    def ip_address(self):
        """Returns public IP address of the running instance."""
        if self._instance_config.local_ssh_port:
            return '127.0.0.1'

        instance = self.instance_deployment.get_instance()
        if not instance:
            raise InstanceNotRunningError(self.instance_config.name)

        return instance.public_ip_address

    @property
    def ssh_user(self):
        return 'ubuntu'

    @property
    def ssh_key_path(self):
        return self.instance_deployment.key_pair.key_path



from spotty.deployment.abstract_instance_volume import AbstractInstanceVolume
from spotty.providers.aws.aws_resources.snapshot import Snapshot
from spotty.providers.aws.aws_resources.volume import Volume
from spotty.providers.aws.config.validation import validate_ebs_volume_parameters


class EbsVolume(AbstractInstanceVolume):

    DP_CREATE_SNAPSHOT = 'create_snapshot'
    DP_UPDATE_SNAPSHOT = 'update_snapshot'
    DP_RETAIN = 'retain'
    DP_DELETE = 'delete'

    def __init__(self, ec2, volume_config: dict, project_name: str, instance_name: str):
        self._ec2 = ec2
        self._name = volume_config['name']
        self._params = validate_ebs_volume_parameters(volume_config['parameters'])

        self._project_name = project_name
        self._instance_name = instance_name

    @property
    def title(self):
        return 'EBS volume'

    @property
    def name(self):
        return self._name

    @property
    def size(self) -> int:
        return self._params['size']

    @property
    def deletion_policy(self) -> str:
        return self._params['deletionPolicy']

    @property
    def deletion_policy_title(self) -> str:
        return {
            EbsVolume.DP_CREATE_SNAPSHOT: 'Create Snapshot',
            EbsVolume.DP_UPDATE_SNAPSHOT: 'Update Snapshot',
            EbsVolume.DP_RETAIN: 'Retain Volume',
            EbsVolume.DP_DELETE: 'Delete Volume',
        }[self.deletion_policy]

    @property
    def ec2_volume_name(self) -> str:
        """Returns EBS volume name."""
        volume_name = self._params['volumeName']
        if not volume_name:
            volume_name = '%s-%s-%s' % (self._project_name.lower(), self._instance_name.lower(), self.name.lower())

        return volume_name

    @property
    def mount_dir(self) -> str:
        if self._params['mountDir']:
            mount_dir = self._params['mountDir']
        else:
            mount_dir = '/mnt/%s' % self.ec2_volume_name

        return mount_dir

    def get_ec2_volume(self) -> Volume:
        return Volume.get_by_name(self._ec2, self.ec2_volume_name)

    def get_snapshot(self) -> Snapshot:
        return Snapshot.get_by_name(self._ec2, self.ec2_volume_name)



import boto3
from botocore.exceptions import ClientError, WaiterError
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.aws.aws_resources.stack import Stack
from spotty.providers.aws.deployment.cf_templates.instance_profile_template import prepare_instance_profile_template


def create_or_update_instance_profile(region, output: AbstractOutputWriter):
    """Creates or updates an instance profile.
    It was moved to a separate stack because creating of an instance profile resource takes 2 minutes.
    """
    cf = boto3.client('cloudformation', region_name=region)

    instance_profile_stack_name = 'spotty-instance-profile'
    template = prepare_instance_profile_template()

    stack = Stack.get_by_name(cf, instance_profile_stack_name)
    try:
        if stack:
            _update_stack(cf, stack, template, output)
        else:
            _create_stack(cf, instance_profile_stack_name, template, output)

        stack = Stack.get_by_name(cf, instance_profile_stack_name)
    except WaiterError:
        stack = None

    if not stack or stack.status not in ['CREATE_COMPLETE', 'UPDATE_COMPLETE']:
        raise ValueError('Stack "%s" was not created.\n'
                         'Please, see CloudFormation logs for the details.' % instance_profile_stack_name)

    profile_arn = [row['OutputValue'] for row in stack.outputs if row['OutputKey'] == 'ProfileArn'][0]

    return profile_arn


def _create_stack(cf, stack_name, template: str, output: AbstractOutputWriter):
    output.write('Creating IAM role for the instance...')

    stack = Stack.create_stack(
        cf=cf,
        StackName=stack_name,
        TemplateBody=template,
        Capabilities=['CAPABILITY_IAM'],
        OnFailure='DELETE',
    )

    # wait for the stack to be created
    stack.wait_stack_created(delay=15)


def _update_stack(cf, stack: Stack, template: str, output: AbstractOutputWriter):
    try:
        updated_stack = stack.update_stack(
            cf=cf,
            StackName=stack.name,
            TemplateBody=template,
            Capabilities=['CAPABILITY_IAM'],
        )
    except ClientError as e:
        # the stack was not updated because there are no changes
        updated_stack = None
        error_code = e.response.get('Error', {}).get('Code', 'Unknown')
        if error_code != 'ValidationError':
            raise e

    if updated_stack:
        # wait for the stack to be updated
        output.write('Updating IAM role for the instance...')
        updated_stack.wait_stack_updated(delay=15)



class AmiNotFoundError(Exception):
    def __init__(self, ami_name):
        super().__init__('The AMI "%s" was not found.\n'
                         'Use the "spotty aws create-ami" command to create an AMI with NVIDIA Docker.'
                         % ami_name)



from typing import List
import os
import yaml
from cfn_tools import CfnYamlLoader, CfnYamlDumper
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.deployment.abstract_instance_volume import AbstractInstanceVolume
from spotty.deployment.container_deployment import ContainerDeployment
from spotty.providers.aws.config.instance_config import InstanceConfig
from spotty.providers.aws.deployment.project_resources.ebs_volume import EbsVolume


def prepare_instance_template(instance_config: InstanceConfig, volumes: List[AbstractInstanceVolume],
                              availability_zone: str, container: ContainerDeployment, output: AbstractOutputWriter):
    """Prepares CloudFormation template to run a Spot Instance."""

    # read and update CF template
    with open(os.path.join(os.path.dirname(__file__), 'data', 'instance.yaml')) as f:
        template = yaml.load(f, Loader=CfnYamlLoader)

    # get volume resources and updated availability zone
    volume_resources = _get_volume_resources(volumes, output)

    # add volume resources to the template
    template['Resources'].update(volume_resources)

    # set availability zone
    if availability_zone:
        template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['Placement'] = {
            'AvailabilityZone': availability_zone,
        }
        output.write('- availability zone: %s' % availability_zone)
    else:
        output.write('- availability zone: auto')

    # set subnet
    if instance_config.subnet_id:
        template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['NetworkInterfaces'] = [
            {
                'SubnetId': instance_config.subnet_id,
                'DeviceIndex': 0,
                'Groups': template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'][
                    'SecurityGroupIds'],
            }]
        del template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['SecurityGroupIds']

    # add ports to the security group
    for port in container.config.ports:
        if port != 22:
            template['Resources']['InstanceSecurityGroup']['Properties']['SecurityGroupIngress'] += [{
                'CidrIp': '0.0.0.0/0',
                'IpProtocol': 'tcp',
                'FromPort': port,
                'ToPort': port,
            }, {
                'CidrIpv6': '::/0',
                'IpProtocol': 'tcp',
                'FromPort': port,
                'ToPort': port,
            }]

    if instance_config.on_demand:
        # run on-demand instance
        del template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'][
            'InstanceMarketOptions']
        output.write('- on-demand instance')
    else:
        # set maximum price
        if instance_config.max_price:
            template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'] \
                ['InstanceMarketOptions']['SpotOptions']['MaxPrice'] = instance_config.max_price

        output.write('- maximum Spot Instance price: %s'
                     % (('%.04f' % instance_config.max_price) if instance_config.max_price else 'on-demand'))

    # set initial docker commands
    if container.config.commands:
        template['Resources']['InstanceLaunchTemplate']['Metadata']['AWS::CloudFormation::Init'] \
            ['docker_container_config']['files']['/tmp/spotty/container/scripts/startup_commands.sh']['content'] \
            = container.config.commands

    return yaml.dump(template, Dumper=CfnYamlDumper)


def _get_volume_attachment_resource(volume_id, device_name):
    attachment_resource = {
        'Type': 'AWS::EC2::VolumeAttachment',
        'Properties': {
            'Device': device_name,
            'InstanceId': {'Ref': 'Instance'},
            'VolumeId': volume_id,
        },
    }

    return attachment_resource


def _get_volume_resource(volume: EbsVolume, output: AbstractOutputWriter):
    # new volume will be created
    volume_resource = {
        'Type': 'AWS::EC2::Volume',
        'DeletionPolicy': 'Retain',
        'Properties': {
            'AvailabilityZone': {'Fn::GetAtt': ['Instance', 'AvailabilityZone']},
            'Tags': [{
                'Key': 'Name',
                'Value': volume.ec2_volume_name,
            }],
        },
    }

    # check if the snapshot exists and restore the volume from it
    snapshot = volume.get_snapshot()
    if snapshot:
        # volume will be restored from the snapshot
        # check size of the volume
        if volume.size and (volume.size < snapshot.size):
            raise ValueError('Specified size for the "%s" volume (%dGB) is less than size of the '
                             'snapshot (%dGB).'
                             % (volume.name, volume.size, snapshot.size))

        # set snapshot ID
        volume_resource['Properties']['SnapshotId'] = snapshot.snapshot_id

        output.write('- volume "%s" will be restored from the snapshot' % volume.ec2_volume_name)

    else:
        # empty volume will be created, check that the size is specified
        if not volume.size:
            raise ValueError('Size for the new volume is required.')

        output.write('- volume "%s" will be created' % volume.ec2_volume_name)

    # set size of the volume
    if volume.size:
        volume_resource['Properties']['Size'] = volume.size

    # set a name for the new volume
    volume_resource['Properties']['Tags'] = [{'Key': 'Name', 'Value': volume.ec2_volume_name}]

    return volume_resource


def _get_volume_resources(volumes: List[AbstractInstanceVolume], output: AbstractOutputWriter):
    resources = {}

    # ending letters for the devices (https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/device_naming.html)
    # TODO: different device names on Nitro-based instances,
    # see: https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/nvme-ebs-volumes.html
    device_letters = 'fghijklmnop'

    # create and attach volumes
    for i, volume in enumerate(volumes):
        if isinstance(volume, EbsVolume):
            device_letter = device_letters[i]

            ec2_volume = volume.get_ec2_volume()
            if ec2_volume:
                # check if the volume is available
                if not ec2_volume.is_available():
                    raise ValueError('EBS volume "%s" is not available.' % volume.ec2_volume_name)

                # check size of the volume
                if volume.size and (volume.size != ec2_volume.size):
                    raise ValueError('Specified size for the "%s" volume (%dGB) doesn\'t match the size of the '
                                     'existing volume (%dGB).' % (volume.name, volume.size, ec2_volume.size))

                output.write('- volume "%s" (%s) will be attached' % (ec2_volume.name, ec2_volume.volume_id))

                volume_id = ec2_volume.volume_id
            else:
                # create Volume resource
                vol_resource_name = 'Volume' + device_letter.upper()
                vol_resource = _get_volume_resource(volume, output)
                resources[vol_resource_name] = vol_resource

                volume_id = {'Ref': vol_resource_name}

            # create VolumeAttachment resource
            vol_attachment_resource_name = 'VolumeAttachment' + device_letter.upper()
            device_name = '/dev/sd' + device_letter
            vol_attachment_resource = _get_volume_attachment_resource(volume_id, device_name)
            resources[vol_attachment_resource_name] = vol_attachment_resource

    return resources



from abc import ABC, abstractmethod
from argparse import Namespace, ArgumentParser
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class AbstractCommand(ABC):
    """Abstract class for a Spotty sub-command."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The sub-command name."""
        raise NotImplementedError

    @property
    def description(self) -> str:
        """The sub-command description. It will be displayed in the help text."""
        return ''

    def configure(self, parser: ArgumentParser):
        """Adds arguments for the sub-command."""
        parser.add_argument('-d', '--debug', action='store_true', help='Show debug messages')

    @abstractmethod
    def run(self, args: Namespace, output: AbstractOutputWriter):
        """Runs the sub-command.

        Args:
            args: Arguments provided by argparse.
            output: Output writer.
        Raises:
            ValueError: If command's arguments can't be processed.
        """
        raise NotImplementedError



from collections import namedtuple, OrderedDict
import os
from subprocess import list2cmdline
from typing import List
from spotty.config.container_config import ContainerConfig
from spotty.config.validation import is_subdir
from spotty.deployment.abstract_instance_volume import AbstractInstanceVolume


VolumeMount = namedtuple('VolumeMount', ['name', 'host_dir', 'container_dir'])


class ContainerDeployment(object):

    def __init__(self, project_name: str, volumes: List[AbstractInstanceVolume], container_config: ContainerConfig):
        self._project_name = project_name
        self._config = container_config

        # get container volumes and host project directory
        self._volume_mounts, self._host_project_dir = self._get_volume_mounts(volumes)

    @property
    def config(self) -> ContainerConfig:
        return self._config

    @property
    def dockerfile_path(self):
        """Dockerfile path on the host OS."""
        dockerfile_path = self.config.file
        if dockerfile_path:
            dockerfile_path = self.host_project_dir + '/' + dockerfile_path

        return dockerfile_path

    @property
    def docker_context_path(self):
        """Docker build's context path on the host OS."""
        dockerfile_path = self.dockerfile_path
        if not dockerfile_path:
            return ''

        return os.path.dirname(dockerfile_path)

    @property
    def host_project_dir(self):
        return self._host_project_dir

    @property
    def volume_mounts(self) -> List[VolumeMount]:
        return self._volume_mounts

    def get_runtime_parameters(self, is_nvidia_runtime: bool):
        """Returns parameters for the ""docker run" command."""
        parameters = self._config.runtime_parameters + ['-td', '--net=host']
        if is_nvidia_runtime:
            parameters += ['--runtime=nvidia']

        for volume_mount in self.volume_mounts:
            parameters += ['-v', '%s:%s' % (volume_mount.host_dir, volume_mount.container_dir)]

        return list2cmdline(parameters)

    def _get_volume_mounts(self, volumes: List[AbstractInstanceVolume]):
        """Get container volume mounts."""
        # get mount directories for the volumes
        mount_dirs = OrderedDict([(volume.name, volume.mount_dir) for volume in volumes])

        # get container volumes mapping
        volume_mounts = []
        for container_volume in self.config.volume_mounts:
            volume_name = container_volume['name']
            host_dir = mount_dirs.get(volume_name, '/tmp/spotty/container/volumes/%s' % volume_name)

            volume_mounts.append(VolumeMount(
                name=volume_name,
                host_dir=host_dir,
                container_dir=container_volume['mountPath'],
            ))

        # get host project directory
        host_project_dir = None
        for name, host_dir, container_dir in volume_mounts:
            if is_subdir(self.config.project_dir, container_dir):
                project_subdir = os.path.relpath(self.config.project_dir, container_dir)
                host_project_dir = host_dir + '/' + project_subdir
                break

        if not host_project_dir:
            # use temporary directory for the project
            host_project_dir = '/tmp/spotty/container/volumes/.project'
            volume_mounts.append(VolumeMount(
                name=None,
                host_dir=host_project_dir,
                container_dir=self.config.project_dir,
            ))

        return volume_mounts, host_project_dir



from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.aws.deployment.abstract_aws_deployment import AbstractAwsDeployment
from spotty.providers.aws.aws_resources.instance import Instance
from spotty.providers.aws.deployment.cf_templates.ami_template import prepare_ami_template
from spotty.providers.aws.deployment.checks import check_az_and_subnet, check_max_price
from spotty.providers.aws.deployment.project_resources.ami_stack import AmiStackResource
from spotty.providers.aws.config.validation import is_gpu_instance


class AmiDeployment(AbstractAwsDeployment):

    # version of the AMI stack
    VERSION = '1.0.4'

    @property
    def ec2_instance_name(self) -> str:
        return 'spotty-ami-%s' % self.instance_config.ami_name.lower()

    @property
    def stack(self):
        return AmiStackResource(self.instance_config.ami_name, self.instance_config.region)

    def deploy(self, debug_mode: bool, output: AbstractOutputWriter):
        # check that the "amiId" parameter is not set
        if self.instance_config.ami_id:
            raise ValueError('The "amiId" parameter cannot be used for creating an AMI.')

        # check that it's a GPU instance type
        instance_type = self.instance_config.instance_type
        if not is_gpu_instance(instance_type):
            raise ValueError('"%s" is not a GPU instance' % instance_type)

        # check that an image with this name doesn't exist yet
        ami = self.get_ami()
        if ami:
            raise ValueError('AMI with the name "%s" already exists.' % ami.name)

        # check availability zone and subnet
        check_az_and_subnet(self._ec2, self.instance_config.region, self.instance_config.availability_zone,
                            self.instance_config.subnet_id)

        # check the maximum price for a spot instance
        availability_zone = self.instance_config.availability_zone
        check_max_price(self._ec2, self.instance_config.instance_type, self.instance_config.on_demand,
                        self.instance_config.max_price, availability_zone)

        # prepare CF template
        subnet_id = self.instance_config.subnet_id
        on_demand = self.instance_config.on_demand
        template = prepare_ami_template(availability_zone, subnet_id, debug_mode, on_demand)

        # create stack
        parameters = self._get_template_parameters(debug_mode)
        self.stack.create_stack(template, parameters, debug_mode, output)

    def _get_template_parameters(self, debug_mode: bool = False):
        parameters = {
            'ImageVersion': self.VERSION,
            'VpcId': self.get_vpc_id(),
            'InstanceType': self.instance_config.instance_type,
            'ImageName': self.instance_config.ami_name,
            'InstanceNameTag': self.ec2_instance_name,
        }

        if debug_mode:
            parameters['DebugMode'] = 'true'
            parameters['KeyName'] = self.key_pair.get_or_create_key()  # get or create a key pair

        return parameters

    def delete(self, output: AbstractOutputWriter):
        # check that the "amiId" parameter is not set
        if self.instance_config.ami_id:
            raise ValueError('The "amiId" parameter cannot be used for deleting an AMI.')

        # check if the AMI stack exists
        stack_id = None
        if not self.stack.get_stack():
            # try to get the stack ID from the AMI tags (for older versions of Spotty)
            ami = self.get_ami()
            if not ami:
                raise ValueError('AMI with the name "%s" not found.' % self.instance_config.ami_name)

            stack_id = ami.get_tag_value('spotty:stack-id')
            if not stack_id:
                raise ValueError('AMI "%s" wasn\'t created by Spotty.' % self.instance_config.ami_name)

        # ask user to confirm the deletion
        confirm = input('AMI "%s" will be deleted.\n'
                        'Type "y" to confirm: ' % self.instance_config.ami_name)
        if confirm != 'y':
            output.write('You didn\'t confirm the operation.')
            return

        self.stack.delete_stack(output, stack_id=stack_id)

    def get_ip_address(self):
        instance = Instance.get_by_stack_name(self._ec2, self.stack.name)
        if not instance:
            return None

        if self._instance_config.local_ssh_port:
            return '127.0.0.1'

        return instance.public_ip_address



from spotty.config.container_config import ContainerConfig
from spotty.config.validation import validate_basic_config
from spotty.utils import filter_list


class ProjectConfig(object):

    def __init__(self, config: dict, project_dir: str):
        # validate the config
        config = validate_basic_config(config, project_dir)

        self._project_dir = project_dir
        self._config = config
        self._container = ContainerConfig(config['container'])

    @property
    def project_dir(self) -> str:
        return self._project_dir

    @property
    def project_name(self) -> str:
        return self._config['project']['name']

    @property
    def sync_filters(self) -> list:
        return self._config['project']['syncFilters']

    @property
    def container(self) -> ContainerConfig:
        return self._container

    @property
    def scripts(self) -> dict:
        return self._config['scripts']

    @property
    def instances(self) -> list:
        return self._config['instances']



class Image(object):

    def __init__(self, ec2, ami_info):
        self._ec2 = ec2
        self._ami_info = ami_info

    @staticmethod
    def get_by_name(ec2, ami_name: str):
        """Returns a AMI by its name."""
        res = ec2.describe_images(Owners=['self'], Filters=[
            {'Name': 'name', 'Values': [ami_name]},
        ])

        if len(res['Images']) > 1:
            raise ValueError('Several AMIs use the same name: "%s".' % ami_name)

        if not len(res['Images']):
            return None

        return Image(ec2, res['Images'][0])

    @staticmethod
    def get_by_id(ec2, ami_id: str):
        """Returns a AMI by its ID."""
        res = ec2.describe_images(Filters=[{'Name': 'image-id', 'Values': [ami_id]}])

        if not len(res['Images']):
            return None

        return Image(ec2, res['Images'][0])

    @property
    def image_id(self) -> str:
        return self._ami_info['ImageId']

    @property
    def name(self) -> str:
        return self._ami_info['Name']

    @property
    def size(self) -> int:
        return self._ami_info['BlockDeviceMappings'][0]['Ebs']['VolumeSize']

    def get_tag_value(self, tag_name):
        tag_values = [tag['Value'] for tag in self._ami_info['Tags'] if tag['Key'] == tag_name]
        if not tag_values:
            return None

        return tag_values[0]



from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class NullOutputWriter(AbstractOutputWriter):

    def _write(self, msg: str):
        """Does nothing."""
        pass



import re
import chevron
from chevron.tokenizer import tokenize


def parse_parameters(script_params: str):
    """Parses script parameters."""
    params = {}
    for param in script_params:
        match = re.match('(\w+)=(.*)', param)
        if not match:
            raise ValueError('Invalid script parameter: "%s"' % param)

        param_name, param_value = match.groups()
        if param_name in params:
            raise ValueError('Parameter "%s" defined twice' % param_name)

        params[param_name] = param_value

    return params


def render_script(template: str, params: dict):
    """Renders a script template.

    It based on the Mustache templates, but only
    variables and delimiter changes are allowed.

    Raises an exception if one of the provided parameters doesn't
    exist in the template.
    """
    tokens = list(tokenize(template))
    template_keys = set()
    for tag, key in tokens:
        if tag not in ['literal', 'no escape', 'variable', 'set delimiter']:
            raise ValueError('Script templates support only variables and delimiter changes.')

        template_keys.add(key)

    # check that the script contains keys for all provided parameters
    for key in params:
        if key not in template_keys:
            raise ValueError('Parameter "%s" doesn\'t exist in the script.' % key)

    return chevron.render(tokens, params)



import boto3
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.aws.aws_resources.stack import Stack


class AmiStackResource(object):

    def __init__(self, ami_name: str, region: str):
        self._cf = boto3.client('cloudformation', region_name=region)
        self._stack_name = 'spotty-ami-%s' % ami_name.lower()

    @property
    def name(self):
        return self._stack_name

    def get_stack(self) -> Stack:
        return Stack.get_by_name(self._cf, self._stack_name)

    def create_stack(self, template: str, parameters: dict, debug_mode: bool, output: AbstractOutputWriter):
        """Creates an AMI stack and waits for the AMI to be created.

        Args:
            template: CloudFormation template
            parameters: parameters for the template
            debug_mode: if "True", NVIDIA Docker will be installed, but an AMI will not be created and the instance
                        will not be terminated, so the user can connect to the instance for debugging.
            output: output writer
        """
        stack = Stack.create_stack(
            cf=self._cf,
            StackName=self._stack_name,
            TemplateBody=template,
            Parameters=[{'ParameterKey': key, 'ParameterValue': value} for key, value in parameters.items()],
            Capabilities=['CAPABILITY_IAM'],
            OnFailure='DO_NOTHING' if debug_mode else 'DELETE',
        )

        output.write('Waiting for the AMI to be created...')

        resource_messages = [
            ('InstanceProfile', 'creating IAM role for the instance'),
            ('Instance', 'launching the instance'),
            ('InstanceReadyWaitCondition', 'installing NVIDIA Docker'),
            ('AMICreatedWaitCondition', 'creating AMI and terminating the instance'),
        ]

        # wait for the stack to be created
        with output.prefix('  '):
            stack = stack.wait_status_changed(waiting_status='CREATE_IN_PROGRESS',
                                              resource_messages=resource_messages,
                                              resource_success_status='CREATE_COMPLETE', output=output)

        if stack.status != 'CREATE_COMPLETE':
            raise ValueError('Stack "%s" was not created.\n'
                             'Please, see CloudFormation logs for the details.' % self._stack_name)

        if debug_mode:
            output.write('Stack "%s" was created in debug mode.' % self._stack_name)
        else:
            ami_id = [row['OutputValue'] for row in stack.outputs if row['OutputKey'] == 'NewAMI'][0]
            output.write('\n'
                         '--------------------------------------------------\n'
                         'AMI "%s" (ID=%s) was successfully created.\n'
                         'Use the "spotty start" command to run an instance.\n'
                         '--------------------------------------------------'
                         % (parameters['ImageName'], ami_id))

    def delete_stack(self, output: AbstractOutputWriter, stack_id=None):
        """Deletes an AMI stack.

        Args:
            output: output writer
            stack_id: ID of the stack to delete (for older versions of Spotty)
        """
        # delete the image
        stack = Stack.get_by_name(self._cf, stack_id) if stack_id else self.get_stack()
        stack.delete()

        output.write('Waiting for the AMI to be deleted...')

        # wait for the deletion to be completed
        with output.prefix('  '):
            stack = stack.wait_status_changed(waiting_status='DELETE_IN_PROGRESS',
                                              resource_messages=[],
                                              resource_success_status='DELETE_COMPLETE', output=output)

        if stack.status == 'DELETE_COMPLETE':
            output.write('\n'
                         '-----------------------------\n'
                         'AMI was successfully deleted.\n'
                         '-----------------------------')
        else:
            raise ValueError('Stack "%s" not deleted.\n'
                             'See CloudFormation and CloudWatch logs for details.' % stack_id)



from spotty.providers.aws.aws_resources.snapshot import Snapshot


class Volume(object):

    def __init__(self, ec2, volume_info):
        self._ec2 = ec2
        self._volume_info = volume_info

    @staticmethod
    def get_by_name(ec2, volume_name: str):
        """Returns a volume by its name."""
        res = ec2.describe_volumes(Filters=[
            {'Name': 'tag:Name', 'Values': [volume_name]},
        ])

        if len(res['Volumes']) > 1:
            raise ValueError('Several volumes with Name=%s found.' % volume_name)

        if not len(res['Volumes']):
            return None

        return Volume(ec2, res['Volumes'][0])

    @property
    def name(self) -> str:
        volume_name = [tag['Value'] for tag in self._volume_info['Tags'] if tag['Key'] == 'Name']
        if not volume_name:
            return ''

        return volume_name[0]

    @property
    def volume_id(self) -> str:
        return self._volume_info['VolumeId']

    @property
    def size(self) -> int:
        return self._volume_info['Size']

    @property
    def availability_zone(self) -> str:
        return self._volume_info['AvailabilityZone']

    def is_available(self):
        return self._volume_info['State'] == 'available'

    def create_snapshot(self) -> Snapshot:
        snapshot_info = self._ec2.create_snapshot(
            VolumeId=self._volume_info['VolumeId'],
            TagSpecifications=[{
                'ResourceType': 'snapshot',
                'Tags': [{
                    'Key': 'Name',
                    'Value': self.name,
                }],
            }],
        )

        return Snapshot(self._ec2, snapshot_info)

    def delete(self):
        return self._ec2.delete_volume(VolumeId=self._volume_info['VolumeId'])



from abc import abstractmethod
from argparse import Namespace, ArgumentParser
import sys
from spotty.commands.abstract_command import AbstractCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.helpers.commands import add_subparsers


class AbstractProviderCommand(AbstractCommand):
    """Abstract class for a provider sub-command."""

    @property
    @abstractmethod
    def commands(self) -> list:
        """Returns a list of the provider sub-commands."""
        raise NotImplementedError

    def configure(self, parser: ArgumentParser):
        add_subparsers(parser, self.commands)

    def run(self, args: Namespace, output: AbstractOutputWriter):
        """If the command is called, it just displays a list of available sub-commands."""
        args.parser.print_help()
        sys.exit(1)



from spotty.config.abstract_instance_config import AbstractInstanceConfig
from spotty.providers.aws.config.validation import validate_instance_parameters

VOLUME_TYPE_EBS = 'ebs'
DEFAULT_AMI_NAME = 'SpottyAMI'


class InstanceConfig(AbstractInstanceConfig):

    def __init__(self, config: dict):
        super().__init__(config)

        self._params = validate_instance_parameters(self._params)

    @property
    def region(self) -> str:
        return self._params['region']

    @property
    def availability_zone(self) -> str:
        return self._params['availabilityZone']

    @property
    def subnet_id(self) -> str:
        return self._params['subnetId']

    @property
    def instance_type(self) -> str:
        return self._params['instanceType']

    @property
    def on_demand(self) -> bool:
        return self._params['onDemandInstance']

    @property
    def ami_name(self) -> str:
        return self._params['amiName'] if self._params['amiName'] else DEFAULT_AMI_NAME

    @property
    def ami_id(self) -> str:
        return self._params['amiId']

    @property
    def root_volume_size(self) -> int:
        return self._params['rootVolumeSize']

    @property
    def max_price(self) -> float:
        return self._params['maxPrice']



from subprocess import list2cmdline
from typing import List
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.config.project_config import ProjectConfig
from spotty.config.validation import is_subdir
from spotty.deployment.abstract_instance_volume import AbstractInstanceVolume
from spotty.providers.aws.config.instance_config import VOLUME_TYPE_EBS
from spotty.deployment.container_deployment import ContainerDeployment, VolumeMount
from spotty.providers.aws.deployment.abstract_aws_deployment import AbstractAwsDeployment
from spotty.providers.aws.deployment.checks import check_az_and_subnet, check_max_price
from spotty.providers.aws.deployment.project_resources.ebs_volume import EbsVolume
from spotty.providers.aws.deployment.cf_templates.instance_template import prepare_instance_template
from spotty.providers.aws.errors.ami_not_found import AmiNotFoundError
from spotty.providers.aws.helpers.download import get_tmp_instance_s3_path
from spotty.providers.aws.helpers.sync import sync_project_with_s3, get_project_s3_path, get_instance_sync_arguments
from spotty.providers.aws.deployment.project_resources.bucket import BucketResource
from spotty.providers.aws.deployment.project_resources.instance_profile_stack import create_or_update_instance_profile
from spotty.providers.aws.deployment.project_resources.instance_stack import InstanceStackResource
from spotty.providers.aws.config.validation import is_nitro_instance, is_gpu_instance
from spotty.utils import render_table


class InstanceDeployment(AbstractAwsDeployment):

    @property
    def ec2_instance_name(self) -> str:
        return '%s-%s' % (self._project_name.lower(), self.instance_config.name.lower())

    @property
    def bucket(self) -> BucketResource:
        return BucketResource(self._project_name, self.instance_config.region)

    @property
    def stack(self) -> InstanceStackResource:
        return InstanceStackResource(self._project_name, self.instance_config.name, self.instance_config.region)

    def get_instance(self):
        return self.stack.get_instance()

    def deploy(self, project_config: ProjectConfig, output: AbstractOutputWriter, dry_run=False):
        # check that it's not a Nitro-based instance
        if is_nitro_instance(self.instance_config.instance_type):
            raise ValueError('Currently Nitro-based instances are not supported.')

        # check availability zone and subnet configuration
        check_az_and_subnet(self._ec2, self.instance_config.region, self.instance_config.availability_zone,
                            self.instance_config.subnet_id)

        # get volumes
        volumes = self._get_volumes()

        # get deployment availability zone
        availability_zone = self._get_availability_zone(volumes)

        # check the maximum price for a spot instance
        check_max_price(self._ec2, self.instance_config.instance_type, self.instance_config.on_demand,
                        self.instance_config.max_price, availability_zone)

        # create or get existing bucket for the project
        bucket_name = self.bucket.get_or_create_bucket(output, dry_run)

        # sync the project with the bucket
        output.write('Syncing the project with S3 bucket...')
        sync_project_with_s3(project_config.project_dir, bucket_name, self.instance_config.region,
                             project_config.sync_filters, dry_run)

        # create or update instance profile
        if not dry_run:
            instance_profile_arn = create_or_update_instance_profile(self.instance_config.region, output)
        else:
            instance_profile_arn = None

        output.write('Preparing CloudFormation template...')

        # prepare CloudFormation template
        container = ContainerDeployment(project_config.project_name, volumes, project_config.container)
        with output.prefix('  '):
            template = prepare_instance_template(self.instance_config, volumes, availability_zone, container,
                                                 output)

            # get parameters for the template
            parameters = self._get_template_parameters(instance_profile_arn, self.instance_config.name, bucket_name,
                                                       project_config.sync_filters, volumes, container, output,
                                                       dry_run=dry_run)

        # print information about the volumes
        output.write('\nVolumes:\n%s\n' % self._render_volumes_info_table(container.volume_mounts, volumes))

        # create stack
        if not dry_run:
            self.stack.create_or_update_stack(template, parameters, output)

    def _get_availability_zone(self, volumes: List[AbstractInstanceVolume]):
        """Checks that existing volumes located in the same AZ and the AZ from the
        config file matches volumes AZ.

        Args:
            volumes: List of volume objects.

        Returns:
            The final AZ where the instance should be run or an empty string if
            the instance can be run in any AZ.

        Raises:
            ValueError: AZ in the config file doesn't match the AZs of the volumes or
                AZs of the volumes are different.
        """
        availability_zone = self.instance_config.availability_zone
        for volume in volumes:
            if isinstance(volume, EbsVolume):
                ec2_volume = volume.get_ec2_volume()
                if ec2_volume:
                    if availability_zone and (availability_zone != ec2_volume.availability_zone):
                        raise ValueError(
                            'The availability zone in the configuration file doesn\'t match the availability zone '
                            'of the existing volume or you have two existing volumes in different availability '
                            'zones.')

                    # update availability zone
                    availability_zone = ec2_volume.availability_zone

        return availability_zone

    def _get_volumes(self) -> List[AbstractInstanceVolume]:
        volumes = []
        for volume_config in self.instance_config.volumes:
            volume_type = volume_config['type']
            if volume_type == VOLUME_TYPE_EBS:
                volumes.append(EbsVolume(self._ec2, volume_config, self._project_name, self.instance_config.name))
            else:
                raise ValueError('AWS volume type "%s" not supported.' % volume_type)

        return volumes

    def _get_template_parameters(self, instance_profile_arn: str, instance_name: str, bucket_name: str,
                                 sync_filters: list, volumes: List[AbstractInstanceVolume],
                                 container: ContainerDeployment, output: AbstractOutputWriter, dry_run=False):
        # get VPC ID
        vpc_id = self.get_vpc_id()

        # get image info
        ami = self.get_ami()
        if not ami:
            raise AmiNotFoundError(self.instance_config.ami_name)

        # check root volume size
        root_volume_size = self.instance_config.root_volume_size
        if root_volume_size and root_volume_size < ami.size:
            raise ValueError('Root volume size cannot be less than the size of AMI (%dGB).' % ami.size)
        elif not root_volume_size:
            # if a root volume size is not specified, make it 5GB larger than the AMI size
            root_volume_size = ami.size + 5

        # create key pair
        key_name = self.key_pair.get_or_create_key(dry_run)

        # get mount directories for the volumes
        mount_dirs = [volume.mount_dir for volume in volumes]

        # get Docker runtime parameters
        runtime_parameters = container.get_runtime_parameters(is_gpu_instance(self.instance_config.instance_type))

        # print info about the Docker data root
        if self.instance_config.docker_data_root:
            docker_data_volume_name = [volume.name for volume in volumes
                                       if is_subdir(self.instance_config.docker_data_root, volume.mount_dir)][0]
            output.write('- Docker data will be stored on the "%s" volume' % docker_data_volume_name)

        # create stack
        parameters = {
            'VpcId': vpc_id,
            'InstanceProfileArn': instance_profile_arn,
            'InstanceType': self.instance_config.instance_type,
            'KeyName': key_name,
            'ImageId': ami.image_id,
            'RootVolumeSize': str(root_volume_size),
            'VolumeMountDirectories': ('"%s"' % '" "'.join(mount_dirs)) if mount_dirs else '',
            'DockerDataRootDirectory': self.instance_config.docker_data_root,
            'DockerImage': container.config.image,
            'DockerfilePath': container.dockerfile_path,
            'DockerBuildContextPath': container.docker_context_path,
            'DockerRuntimeParameters': runtime_parameters,
            'DockerWorkingDirectory': container.config.working_dir,
            'InstanceNameTag': self.ec2_instance_name,
            'ProjectS3Path': get_project_s3_path(bucket_name),
            'HostProjectDirectory': container.host_project_dir,
            'SyncCommandArgs': list2cmdline(get_instance_sync_arguments(sync_filters)),
            'UploadS3Path': get_tmp_instance_s3_path(bucket_name, instance_name),
        }

        return parameters

    def apply_deletion_policies(self, output: AbstractOutputWriter):
        """Applies deletion policies to the EBS volumes."""
        wait_snapshots = []
        delete_snapshots = []
        delete_volumes = []

        # get volumes
        volumes = self._get_volumes()
        ebs_volumes = [volume for volume in volumes if isinstance(volume, EbsVolume)]

        # no volumes
        if not ebs_volumes:
            output.write('- no EBS volumes configured')
            return

        # apply deletion policies
        for volume in ebs_volumes:
            # get EC2 volume
            try:
                ec2_volume = volume.get_ec2_volume()
            except Exception as e:
                output.write('- volume "%s" not found. Error: %s' % (volume.ec2_volume_name, str(e)))
                continue

            if not ec2_volume:
                output.write('- volume "%s" not found' % volume.ec2_volume_name)
                continue

            if not ec2_volume.is_available():
                output.write('- volume "%s" is not available' % volume.ec2_volume_name)
                continue

            # apply deletion policies
            if volume.deletion_policy == EbsVolume.DP_RETAIN:
                # do nothing
                output.write('- volume "%s" is retained' % ec2_volume.name)

            elif volume.deletion_policy == EbsVolume.DP_DELETE:
                # volume will be deleted later
                delete_volumes.append(ec2_volume)

            elif volume.deletion_policy == EbsVolume.DP_CREATE_SNAPSHOT \
                    or volume.deletion_policy == EbsVolume.DP_UPDATE_SNAPSHOT:
                try:
                    # rename or delete previous snapshot
                    prev_snapshot = volume.get_snapshot()
                    if prev_snapshot:
                        # rename previous snapshot
                        if volume.deletion_policy == EbsVolume.DP_CREATE_SNAPSHOT:
                            prev_snapshot.rename('%s-%d' % (prev_snapshot.name, prev_snapshot.creation_time))

                        # once new snapshot will be created, the old one should be deleted
                        if volume.deletion_policy == EbsVolume.DP_UPDATE_SNAPSHOT:
                            delete_snapshots.append(prev_snapshot)

                    output.write('- creating snapshot for the volume "%s"...' % ec2_volume.name)

                    # create new snapshot
                    new_snapshot = ec2_volume.create_snapshot()
                    wait_snapshots.append((new_snapshot, ec2_volume))

                    # once the snapshot will be created, the volume should be deleted
                    delete_volumes.append(ec2_volume)
                except Exception as e:
                    output.write('- snapshot for the volume "%s" was not created. Error: %s'
                                 % (volume.ec2_volume_name, str(e)))

            else:
                raise ValueError('Unsupported deletion policy: "%s".' % volume.deletion_policy)

        # wait until all snapshots will be created
        for snapshot, ec2_volume in wait_snapshots:
            try:
                snapshot.wait_snapshot_completed()
                output.write('- snapshot for the volume "%s" was created' % snapshot.name)
            except Exception as e:
                output.write('- snapshot "%s" was not created. Error: %s' % (snapshot.name, str(e)))

        # delete old snapshots
        for snapshot in delete_snapshots:
            try:
                snapshot.delete()
                output.write('- previous snapshot "%s" was deleted' % snapshot.name)
            except Exception as e:
                output.write('- previous snapshot "%s" was not deleted. Error: %s' % (snapshot.name, str(e)))

        # delete volumes
        for ec2_volume in delete_volumes:
            try:
                ec2_volume.delete()
                output.write('- volume "%s" was deleted' % ec2_volume.name)
            except Exception as e:
                output.write('- volume "%s" was not deleted. Error: %s' % (ec2_volume.name, str(e)))

    @staticmethod
    def _render_volumes_info_table(volume_mounts: List[VolumeMount], volumes: List[AbstractInstanceVolume]):
        table = [('Name', 'Container Dir', 'Type', 'Deletion Policy')]

        # add volume mounts to the info table
        volumes_dict = {volume.name: volume for volume in volumes}
        for volume_mount in volume_mounts:
            if volume_mount.name in volumes_dict:
                # the volume will be mounted to the container
                volume = volumes_dict[volume_mount.name]
                deletion_policy = volume.deletion_policy_title if volume.deletion_policy_title else '-'
                table.append((volume_mount.name, volume_mount.container_dir, volume.title, deletion_policy))
            else:
                # a temporary directory will be mounted to the container
                vol_mount_name = '-' if volume_mount.name is None else volume_mount.name
                table.append((vol_mount_name, volume_mount.container_dir, 'temporary directory', '-'))

        # add volumes that were not mounted to the container to the info table
        volume_mounts_dict = {volume_mount.name for volume_mount in volume_mounts}
        for volume in volumes:
            if volume.name not in volume_mounts_dict:
                deletion_policy = volume.deletion_policy_title if volume.deletion_policy_title else '-'
                table.append((volume.name, '-', volume.title, deletion_policy))

        return render_table(table, separate_title=True)



from abc import ABC, abstractmethod
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.config.abstract_instance_config import AbstractInstanceConfig
from spotty.config.project_config import ProjectConfig


class AbstractInstanceManager(ABC):

    def __init__(self, project_config: ProjectConfig, instance_config: dict):
        self._project_config = project_config
        self._instance_config = self._get_instance_config(instance_config)

    @abstractmethod
    def _get_instance_config(self, config: dict) -> AbstractInstanceConfig:
        """A factory method to create a provider's instance config."""
        raise NotImplementedError

    @abstractmethod
    def is_running(self):
        """Checks if the instance is running."""
        raise NotImplementedError

    @abstractmethod
    def start(self, output: AbstractOutputWriter, dry_run=False):
        """Creates a stack with the instance."""
        raise NotImplementedError

    @abstractmethod
    def stop(self, output: AbstractOutputWriter):
        """Deletes the stack."""
        raise NotImplementedError

    @abstractmethod
    def clean(self, output: AbstractOutputWriter):
        """Deletes the stack."""
        raise NotImplementedError

    @abstractmethod
    def sync(self, output: AbstractOutputWriter, dry_run=False):
        """Synchronizes the project code with the instance."""
        raise NotImplementedError

    @abstractmethod
    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):
        """Downloads files from the instance."""
        raise NotImplementedError

    @property
    @abstractmethod
    def status_text(self) -> str:
        """Information about the running instance that will be
        shown to the user once the instance is started or the "status"
        command is called.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def ip_address(self) -> str:
        """Returns an IP address of the running instance."""
        raise NotImplementedError

    @property
    def ssh_port(self) -> int:
        if self._instance_config.local_ssh_port:
            return self._instance_config.local_ssh_port

        return 22

    @property
    @abstractmethod
    def ssh_user(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def ssh_key_path(self) -> str:
        raise NotImplementedError

    @property
    def project_config(self) -> ProjectConfig:
        return self._project_config

    @property
    def instance_config(self) -> AbstractInstanceConfig:
        return self._instance_config



from spotty.providers.aws.aws_resources.subnet import Subnet
from spotty.providers.aws.helpers.spot_prices import get_current_spot_price


def check_az_and_subnet(ec2, region: str, availability_zone: str, subnet_id: str):
    # get all availability zones for the region
    zones = ec2.describe_availability_zones()
    zone_names = [zone['ZoneName'] for zone in zones['AvailabilityZones']]

    # check availability zone
    if availability_zone and availability_zone not in zone_names:
        raise ValueError('Availability zone "%s" doesn\'t exist in the "%s" region.'
                         % (availability_zone, region))

    if availability_zone:
        if subnet_id:
            subnet = Subnet.get_by_id(ec2, subnet_id)
            if not subnet:
                raise ValueError('Subnet "%s" not found.' % subnet_id)

            if subnet.availability_zone != availability_zone:
                raise ValueError('Availability zone of the subnet doesn\'t match the specified availability zone')
        else:
            default_subnets = Subnet.get_default_subnets(ec2)
            default_subnet = [subnet for subnet in default_subnets
                              if subnet.availability_zone == availability_zone]
            if not default_subnet:
                raise ValueError('Default subnet for the "%s" availability zone not found.\n'
                                 'Use the "subnetId" parameter to specify a subnet for this availability zone.'
                                 % availability_zone)
    else:
        if subnet_id:
            raise ValueError('An availability zone should be specified if a custom subnet is used.')
        else:
            default_subnets = Subnet.get_default_subnets(ec2)
            default_azs = {subnet.availability_zone for subnet in default_subnets}
            zones_wo_subnet = [zone_name for zone_name in zone_names if zone_name not in default_azs]
            if zones_wo_subnet:
                raise ValueError('Default subnets for the following availability zones were not found: %s.\n'
                                 'Use "subnetId" and "availabilityZone" parameters or create missing default '
                                 'subnets.' % ', '.join(zones_wo_subnet))


def check_max_price(ec2, instance_type: str, on_demand: bool, max_price: float, availability_zone: str = ''):
    """Checks that the specified maximum Spot price is less than the
    current Spot price.

    Args:
        ec2: EC2 client
        instance_type (str): Instance Type
        on_demand (bool): True if it's on-demand instance
        max_price (float): requested maximum price for the instance
        availability_zone (str): Availability zone to check. If it's an empty string,
            checks the cheapest AZ.

    Raises:
        ValueError: Current price for the instance is higher than the
            maximum price in the configuration file.
    """
    if not on_demand and max_price:
        current_price = get_current_spot_price(ec2, instance_type, availability_zone)
        if current_price > max_price:
            raise ValueError('Current price for the instance (%.04f) is higher than the maximum price in the '
                             'configuration file (%.04f).' % (current_price, max_price))



from argparse import Namespace, ArgumentParser
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class SyncCommand(AbstractConfigCommand):

    name = 'sync'
    description = 'Synchronize the project with the running instance'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('--dry-run', action='store_true', help='Show files to be synced')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        dry_run = args.dry_run
        with output.prefix('[dry-run] ' if dry_run else ''):
            instance_manager.sync(output, dry_run)

        output.write('Done')



from schema import Schema, Optional, And, Regex, Or, Use
from spotty.config.validation import validate_config, get_instance_parameters_schema


def validate_instance_parameters(params: dict):
    from spotty.providers.aws.config.instance_config import VOLUME_TYPE_EBS

    instance_parameters = {
        'region': And(str, Regex(r'^[a-z0-9-]+$')),
        Optional('availabilityZone', default=''): And(str, Regex(r'^[a-z0-9-]+$')),
        Optional('subnetId', default=''): And(str, Regex(r'^subnet-[a-z0-9]+$')),
        'instanceType': And(str, And(is_valid_instance_type, error='Invalid instance type.')),
        Optional('onDemandInstance', default=False): bool,
        Optional('amiName', default=None): And(str, len, Regex(r'^[\w\(\)\[\]\s\.\/\'@-]{3,128}$')),
        Optional('amiId', default=None): And(str, len, Regex(r'^ami-[a-z0-9]+$')),
        Optional('rootVolumeSize', default=0): And(Or(int, str), Use(str),
                                                   Regex(r'^\d+$', error='Incorrect value for "rootVolumeSize".'),
                                                   Use(int),
                                                   And(lambda x: x > 0,
                                                       error='"rootVolumeSize" should be greater than 0 or should '
                                                             'not be specified.'),
                                                   ),
        Optional('maxPrice', default=0): And(Or(float, int, str), Use(str),
                                             Regex(r'^\d+(\.\d{1,6})?$', error='Incorrect value for "maxPrice".'),
                                             Use(float),
                                             And(lambda x: x > 0, error='"maxPrice" should be greater than 0 or '
                                                                        'should  not be specified.'),
                                             ),
    }

    volumes_checks = [
        And(lambda x: len(x) < 12, error='Maximum 11 volumes are supported at the moment.'),
    ]

    instance_checks = [
        And(lambda x: not (x['onDemandInstance'] and x['maxPrice']),
            error='"maxPrice" cannot be specified for on-demand instances'),
        And(lambda x: not (x['amiName'] and x['amiId']),
            error='"amiName" and "amiId" parameters cannot be used together'),
    ]

    schema = get_instance_parameters_schema(instance_parameters, VOLUME_TYPE_EBS, instance_checks, volumes_checks)

    return validate_config(schema, params)


def validate_ebs_volume_parameters(params: dict):
    from spotty.providers.aws.deployment.project_resources.ebs_volume import EbsVolume

    schema = Schema({
        Optional('volumeName', default=''): And(str, Regex(r'^[\w-]{1,255}$')),
        Optional('mountDir', default=''): str,  # all the checks happened in the base configuration
        Optional('size', default=0): And(int, lambda x: x > 0),
        Optional('deletionPolicy', default=EbsVolume.DP_CREATE_SNAPSHOT): And(
            str,
            lambda x: x in [EbsVolume.DP_CREATE_SNAPSHOT,
                            EbsVolume.DP_UPDATE_SNAPSHOT,
                            EbsVolume.DP_RETAIN,
                            EbsVolume.DP_DELETE], error='Incorrect value for "deletionPolicy".'
        ),
    })

    return validate_config(schema, params)


def is_gpu_instance(instance_type: str):
    return instance_type in [
        'p2.xlarge', 'p2.8xlarge', 'p2.16xlarge',
        'p3.2xlarge', 'p3.8xlarge', 'p3.16xlarge',
        'g3s.xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3.16xlarge',
    ]


def is_nitro_instance(instance_type):
    return instance_type in [
        'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.18xlarge',
        'c5d.large', 'c5d.xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.18xlarge',
        'm5.large', 'm5.xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.12xlarge', 'm5.24xlarge',
        'm5d.large', 'm5d.xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.12xlarge', 'm5d.24xlarge',
        'i3.metal',
    ]


def is_valid_instance_type(instance_type: str):
    return instance_type in [
        't1.micro',
        't2.nano', 't2.micro', 't2.small', 't2.medium', 't2.large', 't2.xlarge', 't2.2xlarge',
        'm1.small', 'm1.medium', 'm1.large', 'm1.xlarge',
        'm3.medium', 'm3.large', 'm3.xlarge', 'm3.2xlarge',
        'm4.large', 'm4.xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.10xlarge', 'm4.16xlarge',
        'm2.xlarge', 'm2.2xlarge', 'm2.4xlarge',
        'cr1.8xlarge',
        'r3.large', 'r3.xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge',
        'r4.large', 'r4.xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.16xlarge',
        'x1.16xlarge', 'x1.32xlarge',
        'x1e.xlarge', 'x1e.2xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.16xlarge', 'x1e.32xlarge',
        'i2.xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge',
        'i3.large', 'i3.xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.16xlarge', 'i3.metal',
        'hi1.4xlarge',
        'hs1.8xlarge',
        'c1.medium', 'c1.xlarge',
        'c3.large', 'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge',
        'c4.large', 'c4.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge',
        'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.18xlarge',
        'c5d.large', 'c5d.xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.18xlarge',
        'cc1.4xlarge',
        'cc2.8xlarge',
        'g2.2xlarge', 'g2.8xlarge',
        'g3s.xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3.16xlarge',
        'cg1.4xlarge',
        'p2.xlarge', 'p2.8xlarge', 'p2.16xlarge',
        'p3.2xlarge', 'p3.8xlarge', 'p3.16xlarge',
        'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge',
        'f1.2xlarge', 'f1.16xlarge',
        'm5.large', 'm5.xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.12xlarge', 'm5.24xlarge',
        'm5d.large', 'm5d.xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.12xlarge', 'm5d.24xlarge',
        'h1.2xlarge', 'h1.4xlarge', 'h1.8xlarge', 'h1.16xlarge',
    ]



import logging
import os
from typing import List
from schema import Schema, And, Use, Optional, Or, Regex, SchemaError


def validate_basic_config(data, project_dir):
    is_old_config = 'instance' in data and 'container' not in data
    if is_old_config:
        logging.warning('The format of the configuration file that you\'re using is deprecated and won\'t be '
                        'supported in the future versions of Spotty.')

        data = convert_old_config(validate_old_config(data, project_dir))

    schema = Schema({
        'project': {
            'name': And(str, Regex(r'^[a-zA-Z0-9][a-zA-Z0-9-]{,26}[a-zA-Z0-9]$')),
            Optional('syncFilters', default=[]): [And(
                {
                    Optional('exclude'): [And(str, len)],
                    Optional('include'): [And(str, len)],
                },
                And(lambda x: x, error='Either "exclude" or "include" filter should be specified.'),
                And(lambda x: not ('exclude' in x and 'include' in x), error='"exclude" and "include" filters should '
                                                                             'be specified as different list items.'),
            )]
        },
        'container': And(
            {
                'projectDir': And(str,
                                  And(os.path.isabs,
                                      error='Use an absolute path when specifying the project directory'),
                                  Use(lambda x: x.rstrip('/'))
                                  ),
                Optional('image', default=''): And(str, len),
                Optional('file', default=''): And(str,  # TODO: a proper regex that the filename is valid
                                                  Regex(r'^[\w\.\/@-]*$',
                                                        error='Invalid name for a Dockerfile'),
                                                  And(lambda x: not x.endswith('/'),
                                                      error='Invalid name for a Dockerfile'),
                                                  And(lambda x: not os.path.isabs(x),
                                                      error='Path to the Dockerfile should be relative to the '
                                                            'project\'s root directory.'),
                                                  And(lambda x: os.path.isfile(os.path.join(project_dir, x)),
                                                      error='Dockerfile not found.'),
                                                  ),
                Optional('volumeMounts', default=[]): (And(
                    [{
                        'name': And(Or(int, str), Use(str), Regex(r'^[\w-]+$')),
                        'mountPath': And(
                            str,
                            And(os.path.isabs, error='Use an absolute path when specifying a mount directory'),
                            Use(lambda x: x.rstrip('/')),
                        ),
                    }],
                    And(lambda x: is_unique_value(x, 'name'),
                        error='Each volume mount must have a unique name.'),
                    And(lambda x: not has_prefix([(volume['mountPath'] + '/') for volume in x]),
                        error='Volume mount paths cannot be prefixes for each other.'),
                )),
                Optional('workingDir', default=''): And(str,
                                                        And(os.path.isabs,
                                                            error='Use an absolute path when specifying a '
                                                                  'working directory'),
                                                        Use(lambda x: x.rstrip('/'))
                                                        ),
                Optional('commands', default=''): str,
                Optional('ports', default=[]): [And(int, lambda x: 0 <= x <= 65535)],
                Optional('runtimeParameters', default=[]): [str],
            },
            And(lambda x: x['image'] or x['file'], error='Either "image" or "file" should be specified.'),
            And(lambda x: not (x['image'] and x['file']), error='"image" and "file" cannot be specified together.'),
        ),
        'instances': And(
            [{
                'name': And(Or(int, str), Use(str), Regex(r'^[\w-]+$')),
                'provider': str,
                'parameters': And({
                    And(str, Regex(r'^[\w]+$')): object,
                }, error='Instance parameters are not specified')
            }],
            And(lambda x: is_unique_value(x, 'name'), error='Each instance must have a unique name.'),
        ),
        Optional('scripts', default={}): {
            And(str, Regex(r'^[\w-]+$')): And(str, len),
        },
    })

    return validate_config(schema, data)


def get_instance_parameters_schema(instance_parameters: dict, default_volume_type: str,
                                   instance_checks: list = None, volumes_checks: list = None):
    if not instance_checks:
        instance_checks = []

    if not volumes_checks:
        volumes_checks = []

    schema = Schema(And(
        {
            **instance_parameters,
            Optional('dockerDataRoot', default=''): And(
                str,
                And(os.path.isabs, error='Use an absolute path when specifying a Docker data root directory'),
                Use(lambda x: x.rstrip('/')),
            ),
            Optional('volumes', default=[]): And(
                [{
                    'name': And(Or(int, str), Use(str), Regex(r'^[\w-]+$')),
                    Optional('type', default=default_volume_type): str,
                    Optional('parameters', default={}): {
                        Optional('mountDir', default=''): And(
                            str,
                            And(os.path.isabs, error='Use absolute paths for mount directories'),
                            Use(lambda x: x.rstrip('/'))
                        ),
                        And(str, Regex(r'^[\w]+$')): object,
                    },
                }],
                And(lambda x: is_unique_value(x, 'name'), error='Each instance volume must have a unique name.'),
                And(lambda x: not has_prefix([(volume['parameters']['mountDir'] + '/') for volume in x
                                              if volume['parameters']['mountDir']]),
                    error='Mount directories cannot be prefixes for each other.'),
                *volumes_checks,
            ),
            Optional('localSshPort', default=None): Or(None, And(int, lambda x: 0 <= x <= 65535)),
        },
        And(lambda x: not x['dockerDataRoot'] or
                      [True for v in x['volumes'] if v['parameters']['mountDir'] and
                       is_subdir(x['dockerDataRoot'], v['parameters']['mountDir'])],
            error='The "mountDir" of one of the volumes must be a prefix for the "dockerDataRoot" path.'),
        *instance_checks
    ))

    return schema


def validate_old_config(data, project_dir):
    schema = Schema({
        'project': {
            'name': And(str, Regex(r'^[a-zA-Z0-9][a-zA-Z0-9-]{,26}[a-zA-Z0-9]$')),
            'remoteDir': And(str,
                             And(os.path.isabs,
                                 error='Use an absolute path when specifying a remote directory'),
                             Use(lambda x: x.rstrip('/'))
                             ),
            Optional('syncFilters', default=[]): [And(
                {
                    Optional('exclude'): [And(str, len)],
                    Optional('include'): [And(str, len)],
                },
                And(lambda x: x, error='Either "exclude" or "include" filter should be specified.'),
                And(lambda x: not ('exclude' in x and 'include' in x), error='"exclude" and "include" filters should '
                                                                             'be specified as different list items.'),
            )]
        },
        'instance': {
            'region': And(str, len),
            Optional('availabilityZone', default=''): str,
            Optional('subnetId', default=''): str,
            'instanceType': str,
            Optional('amiName', default='SpottyAMI'): str,
            Optional('keyName', default=''): str,
            Optional('rootVolumeSize', default=0): And(Or(int, str), Use(str),
                                                       Regex(r'^\d+$', error='Incorrect value for "rootVolumeSize".'),
                                                       Use(int),
                                                       And(lambda x: x > 0,
                                                           error='"rootVolumeSize" should be greater than 0 or should '
                                                                 ' not be specified.'),
                                                       ),
            Optional('maxPrice', default=0): And(Or(float, int, str), Use(str),
                                                 Regex(r'^\d+(\.\d{1,6})?$', error='Incorrect value for "maxPrice".'),
                                                 Use(float),
                                                 And(lambda x: x > 0, error='"maxPrice" should be greater than 0 or '
                                                                            'should  not be specified.'),
                                                 ),
            Optional('volumes', default=[]): And(
                [{
                    Optional('name', default=''): And(str, Regex(r'^[\w-]{1,255}$')),
                    'directory': And(str, lambda x: x.startswith('/'), Use(lambda x: x.rstrip('/'))),
                    Optional('size', default=0): And(int, lambda x: x > 0),
                    Optional('deletionPolicy',
                             default='create_snapshot'): And(str, lambda x: x in ['create_snapshot', 'update_snapshot',
                                                                                  'retain', 'delete'],
                                                             error='Incorrect value for "deletionPolicy".'
                                                             ),
                }],
                And(lambda x: len(x) < 12, error='Maximum 11 volumes are supported at the moment.'),
            ),
            'docker': And(
                {
                    Optional('image', default=''): str,
                    Optional('file', default=''): And(str,  # TODO: a proper regex that the filename is valid
                                                      Regex(r'^[\w\.\/@-]*$',
                                                            error='Invalid name for a Dockerfile'),
                                                      And(lambda x: not x.endswith('/'),
                                                          error='Invalid name for a Dockerfile'),
                                                      And(lambda x: not os.path.isabs(x),
                                                          error='Path to the Dockerfile should be relative to the '
                                                                'project\'s root directory.'),
                                                      And(lambda x: os.path.isfile(os.path.join(project_dir, x)),
                                                          error='Dockerfile not found.'),
                                                      ),
                    Optional('workingDir', default=''): And(str,
                                                            And(os.path.isabs,
                                                                error='Use an absolute path when specifying a '
                                                                      'working directory'),
                                                            ),
                    Optional('dataRoot', default=''): And(str,
                                                          And(os.path.isabs,
                                                              error='Use an absolute path when specifying a Docker '
                                                                    'data root directory'),
                                                          Use(lambda x: x.rstrip('/')),
                                                          ),
                    Optional('commands', default=''): str,
                    Optional('runtimeParameters', default=[]): [str],
                },
                And(lambda x: x['image'] or x['file'], error='Either "image" or "file" should be specified.'),
                And(lambda x: not (x['image'] and x['file']), error='"image" and "file" cannot be specified together.'),
            ),
            Optional('ports', default=[]): [And(int, lambda x: 0 <= x <= 65535)],
            Optional('localSshPort', default=None): And(int, lambda x: 0 <= x <= 65535),
        },
        Optional('scripts', default={}): {
            And(str, Regex(r'^[\w-]+$')): And(str, len),
        },
    })

    return validate_config(schema, data)


def convert_old_config(config):
    from spotty.providers.aws.config.instance_config import VOLUME_TYPE_EBS

    new_config = {
        'project': {
            'name': config['project']['name'],
            'syncFilters': config['project']['syncFilters'],
        },
        'container': {
            'projectDir': config['project']['remoteDir'],
            'image': config['instance']['docker']['image'],
            'file': config['instance']['docker']['file'],
            'volumeMounts': [{
                'name': volume['name'],
                'mountPath': volume['directory'],
            } for volume in config['instance']['volumes']],
            'workingDir': config['instance']['docker']['workingDir'],
            'commands': config['instance']['docker']['commands'],
            'ports': config['instance']['ports'],
            'runtimeParameters': config['instance']['docker']['runtimeParameters'],
        },
        'instances': [{
            'name': 'instance',
            'provider': 'aws',
            'parameters': {
                'region': config['instance']['region'],
                'availabilityZone': config['instance']['availabilityZone'],
                'subnetId': config['instance']['subnetId'],
                'instanceType': config['instance']['instanceType'],
                'amiName': config['instance']['amiName'],
                'rootVolumeSize': config['instance']['rootVolumeSize'],
                'dockerDataRoot': config['instance']['docker']['dataRoot'],
                'maxPrice': config['instance']['maxPrice'],
                'localSshPort': config['instance']['localSshPort'],
                'volumes': [{
                    'name': volume['name'],
                    'type': VOLUME_TYPE_EBS,
                    'parameters': {
                        'volumeName': volume['name'],
                        'mountDir': volume['directory'],
                        'size': volume['size'],
                        'deletionPolicy': volume['deletionPolicy'],
                    },
                } for volume in config['instance']['volumes']],
            },
        }],
        'scripts': config['scripts'],
    }

    def clear_empty_values(c):
        if isinstance(c, dict):
            c = {key: clear_empty_values(c[key]) for key in c if c[key]}
        elif isinstance(c, list):
            c = [clear_empty_values(row) for row in c]

        return c

    new_config = clear_empty_values(new_config)

    return new_config


def is_unique_value(x: List[dict], key):
    """Returns "True" if all values of the key in the list of dictionaries are unique."""
    return len(x) == len(set([v[key] for v in x]))


def has_prefix(x: list):
    """Returns "True" if some value in the list is a prefix for another value in this list."""
    for val in x:
        if len(list(filter(val.startswith, x))) > 1:
            return True

    return False


def is_subdir(subdir_path, dir_path):
    """Returns "True" if it's the second path parameter is a subdirectory of the first path parameter."""
    return (subdir_path.rstrip('/') + '/').startswith(dir_path.rstrip('/') + '/')


def validate_config(schema: Schema, config):
    try:
        validated = schema.validate(config)
    except SchemaError as e:
        raise ValueError(e.errors[-1] if e.errors[-1] else e.autos[-1])

    return validated



import os
import yaml
from cfn_tools import CfnYamlDumper, CfnYamlLoader


def prepare_ami_template(availability_zone: str, subnet_id: str, debug_mode: bool = False, on_demand: bool = False):
    """Prepares CloudFormation template to run a Spot Instance."""

    # read and update CF template
    with open(os.path.join(os.path.dirname(__file__), 'data', 'ami.yaml')) as f:
        template = yaml.load(f, Loader=CfnYamlLoader)

    # remove the key parameter if it's not a debug mode
    if not debug_mode:
        del template['Parameters']['KeyName']
        del template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['KeyName']

    # set availability zone
    if availability_zone:
        template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['Placement'] = {
            'AvailabilityZone': availability_zone,
        }

    # set subnet
    if subnet_id:
        template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['NetworkInterfaces'] = [
            {
                'SubnetId': subnet_id,
                'DeviceIndex': 0,
            }]

    # run on-demand instance
    if on_demand:
        del template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'][
            'InstanceMarketOptions']

    return yaml.dump(template, Dumper=CfnYamlDumper)



class ContainerConfig(object):

    def __init__(self, container_config: dict):
        self._config = container_config

    @property
    def project_dir(self):
        return self._config['projectDir']

    @property
    def image(self):
        return self._config['image']

    @property
    def file(self):
        return self._config['file']

    @property
    def volume_mounts(self):
        return self._config['volumeMounts']

    @property
    def ports(self):
        return list(set(self._config['ports']))

    @property
    def commands(self):
        return self._config['commands']

    @property
    def working_dir(self):
        """Working directory for the Docker container."""
        working_dir = self._config['workingDir']
        if not working_dir:
            working_dir = self._config['projectDir']

        return working_dir

    @property
    def runtime_parameters(self):
        return self._config['runtimeParameters']



from abc import ABC, abstractmethod


class AbstractInstanceVolume(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        """Unique name of the volume that will be used for the deployment."""
        raise NotImplementedError

    @property
    @abstractmethod
    def mount_dir(self) -> str:
        """A directory where the volume should be mounted on the host OS."""
        raise NotImplementedError

    @property
    @abstractmethod
    def title(self) -> str:
        """A title for the volume type.
        It will be used to display information about the volumes during the deployment.
        """
        raise NotImplementedError

    @property
    def deletion_policy_title(self) -> str:
        """A title for the volume's deletion policy.
        It will be used to display information about the volumes during the deployment.
        """
        return ''



from abc import ABC


class AbstractInstanceConfig(ABC):

    def __init__(self, config: dict):
        self._name = config['name']
        self._params = config['parameters']

    @property
    def name(self) -> str:
        """Name of the instance."""
        return self._name

    @property
    def volumes(self) -> list:
        """List of volume configs."""
        return self._params['volumes']

    @property
    def docker_data_root(self) -> str:
        """Data root directory for Docker daemon."""
        return self._params['dockerDataRoot']

    @property
    def local_ssh_port(self) -> int:
        """Local SSH port to connect to the instance (in case of a tunnel)."""
        return self._params['localSshPort']



from argparse import Namespace, ArgumentParser
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.abstract_instance_manager import AbstractInstanceManager


class DownloadCommand(AbstractConfigCommand):

    name = 'download'
    description = 'Download files from the running instance'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-f', '--filters', metavar='FILTER', nargs='+', type=str, required=True,
                            help='AWS S3 Sync "include" filters')
        parser.add_argument('--dry-run', action='store_true', help='Show files to be downloaded')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        filters = [
            {'exclude': ['*']},
            {'include': args.filters}
        ]

        dry_run = args.dry_run
        with output.prefix('[dry-run] ' if dry_run else ''):
            instance_manager.download(filters, output, dry_run)

        output.write('Done')



class Vpc(object):

    def __init__(self, ec2, vpc_info):
        self._ec2 = ec2
        self._vpc_info = vpc_info

    @staticmethod
    def get_default_vpc(ec2):
        """Returns a default VPC."""
        res = ec2.describe_vpcs(Filters=[{'Name': 'isDefault', 'Values': ['true']}])
        if not len(res['Vpcs']):
            return None

        return Vpc(ec2, res['Vpcs'][0])

    @property
    def vpc_id(self) -> str:
        return self._vpc_info['VpcId']



from spotty.deployment.abstract_instance_volume import AbstractInstanceVolume


class TestVolume(AbstractInstanceVolume):

    def __init__(self, volume_name, mount_dir):
        self._volume_name = volume_name
        self._mount_dir = mount_dir

    @property
    def title(self) -> str:
        return 'test'

    @property
    def name(self) -> str:
        return self._volume_name

    @property
    def mount_dir(self) -> str:
        return self._mount_dir



from abc import abstractmethod
import yaml
import os
from argparse import Namespace, ArgumentParser
from spotty.config.project_config import ProjectConfig
from spotty.providers.abstract_instance_manager import AbstractInstanceManager
from spotty.providers.instance_manager_factory import InstanceManagerFactory
from spotty.utils import filter_list
from yaml.scanner import ScannerError
from spotty.commands.abstract_command import AbstractCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class AbstractConfigCommand(AbstractCommand):
    """Abstract class for a Spotty sub-command that needs to use a project's configuration."""

    @abstractmethod
    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        raise NotImplementedError

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-c', '--config', type=str, default=None, help='Path to the configuration file')
        parser.add_argument('instance_name', metavar='INSTANCE_NAME', nargs='?', type=str, help='Instance name')

    def run(self, args: Namespace, output: AbstractOutputWriter):
        # get project directory
        config_path = args.config
        if not config_path:
            config_path = 'spotty.yaml'

        if os.path.isabs(config_path):
            config_abs_path = config_path
        else:
            config_abs_path = os.path.abspath(os.path.join(os.getcwd(), config_path))

        if not os.path.exists(config_abs_path):
            raise ValueError('Configuration file "%s" not found.' % config_path)

        # get project configuration
        project_dir = os.path.dirname(config_abs_path)
        project_config = ProjectConfig(self._load_config(config_abs_path), project_dir)

        # get instance configuration
        instance_config = self._get_instance_config(project_config, args.instance_name, output)

        # create an instance manger
        instance_manager = InstanceManagerFactory.get_instance(project_config, instance_config)

        # run the command
        self._run(instance_manager, args, output)

    @staticmethod
    def _load_config(config_path: str):
        """Returns project configuration."""
        config = None
        if os.path.exists(config_path):
            with open(config_path, 'r') as f:
                try:
                    config = yaml.safe_load(f)
                except ScannerError as e:
                    raise ValueError(str(e))

        return config

    @staticmethod
    def _get_instance_config(project_config: ProjectConfig, instance_name: str, output: AbstractOutputWriter):
        if not instance_name:
            if len(project_config.instances) > 1:
                # ask user to choose the instance
                output.write('Select the instance:\n')
                with output.prefix('  '):
                    for i, instance_config in enumerate(project_config.instances):
                        output.write('[%d] %s' % (i + 1, instance_config['name']))
                output.write()

                try:
                    num = int(input('Enter number: '))
                    output.write()
                except ValueError:
                    num = 0

                if num < 1 or num > len(project_config.instances):
                    raise ValueError('The value from 1 to %d was expected.' % len(project_config.instances))
            else:
                num = 1

            instance_config = project_config.instances[num - 1]
        else:
            # get the instance by name
            instance_configs = filter_list(project_config.instances, 'name', instance_name)
            if not instance_configs:
                raise ValueError('Instance "%s" not found in the configuration file' % instance_name)

            instance_config = instance_configs[0]

        return instance_config



from argparse import Namespace, ArgumentParser
import subprocess
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.helpers.ssh import get_ssh_command
from spotty.providers.abstract_instance_manager import AbstractInstanceManager
from spotty.providers.aws.instance_manager import InstanceManager


class CreateAmiCommand(AbstractConfigCommand):

    name = 'create-ami'
    description = 'Create AMI with NVIDIA Docker'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('--debug-mode', action='store_true', help='Don\'t terminate the instance on failure')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that it's an AWS instance
        if not isinstance(instance_manager, InstanceManager):
            raise ValueError('Instance "%s" is not an AWS instance.' % instance_manager.instance_config.name)

        deployment = instance_manager.ami_deployment

        try:
            deployment.deploy(args.debug_mode, output)
        finally:
            if args.debug_mode:
                ip_address = deployment.get_ip_address()
                if ip_address:
                    ssh_command = get_ssh_command(ip_address, instance_manager.ssh_port, instance_manager.ssh_user,
                                                  instance_manager.ssh_key_path, 'tmux')

                    output.write('\nUse the following command to connect to the instance:\n'
                                 '  %s\n' % subprocess.list2cmdline(ssh_command))



import subprocess
from spotty.helpers.ssh import get_ssh_command
from spotty.providers.aws.helpers.aws_cli import AwsCli


def get_project_s3_path(bucket_name: str):
    return 's3://%s/project' % bucket_name


def get_instance_sync_arguments(sync_filters: list):
    return AwsCli.get_s3_sync_arguments(sync_filters, exact_timestamp=True, quote=True)


def sync_project_with_s3(project_dir, bucket_name, region, sync_filters, dry_run=False):
    # sync the project with S3, deleted files will be deleted from S3
    AwsCli(region=region).s3_sync(project_dir, get_project_s3_path(bucket_name), filters=sync_filters, delete=True,
                                  capture_output=False, dry_run=dry_run)


def sync_instance_with_s3(sync_filters: list, host: str, port: int, user: str, key_path: str):
    """Syncs the project from the S3 bucket to the instance."""

    # "sudo" should be called with the "-i" flag to use the root environment, so aws-cli will read
    # the config file from the root home directory
    remote_cmd = subprocess.list2cmdline(['sudo', '-i', '/tmp/spotty/instance/scripts/sync_project.sh',
                                          *get_instance_sync_arguments(sync_filters), '>', '/dev/null'])

    # connect to the instance and run remote command
    ssh_command = get_ssh_command(host, port, user, key_path, remote_cmd, quiet=True)
    subprocess.call(ssh_command)
